"""Base class for the ``sim.oper`` object"""

import shutil
from subprocess import PIPE, run

from fluidsimfoam.foam_input_files.fields import VolVectorField


class Operators:
    def __init__(self, sim):
        self.sim = sim

        if hasattr(sim.output.input_files, "block_mesh_dict"):
            assert (sim.output.path_run / "system/blockMeshDict").exists()

    def get_cells_coords(self):
        path_c = self.sim.path_run / "0/C"

        if not path_c.exists():
            path_polymesh = self.sim.path_run / "constant/polyMesh/points"

            if not path_polymesh.exists():
                self.sim.make.exec("polymesh", stdout=PIPE)

            if not path_polymesh.exists():
                raise RuntimeError(f"{path_polymesh} does not exists")

            path_postProcess = shutil.which("postProcess")

            if path_postProcess is None:
                raise RuntimeError("OpenFOAM not available")

            run(
                ["postProcess", "-func", "writeCellCentres"],
                cwd=self.sim.path_run,
                stdout=PIPE,
            )

        field = VolVectorField.from_path(path_c, skip_boundary_field=True)
        return field.get_components()



"""Base class for ``Simul.InfoSolver``"""

from fluidsim_core.info import InfoSolverCore


class InfoSolver(InfoSolverCore):
    """Contains the information on which classes are used in a solver"""

    def _init_root(self):
        super()._init_root()

        self.classes._set_child(
            "Output",
            attribs={
                "module_name": "fluidsimfoam.output",
                "class_name": "Output",
            },
        )
        self.classes._set_child(
            "Oper",
            attribs={
                "module_name": "fluidsimfoam.operators",
                "class_name": "Operators",
            },
        )
        self.classes._set_child(
            "InitFields",
            attribs={
                "module_name": "fluidsimfoam.init_fields",
                "class_name": "InitFields",
            },
        )
        self.classes._set_child(
            "Make",
            attribs={
                "module_name": "fluidsimfoam.make",
                "class_name": "MakeInvoke",
            },
        )

    def complete_with_classes(self):
        """Populate info solver by executing ``_complete_info_solver`` class
        methods
        """
        dict_classes = self.import_classes()
        for Class in list(dict_classes.values()):
            if hasattr(Class, "_complete_info_solver"):
                Class._complete_info_solver(self)



"""Helper to create field files

"""

import re
from abc import ABC, abstractmethod
from io import StringIO
from numbers import Number
from pathlib import Path

import numpy as np

from fluidsimfoam.foam_input_files import parse
from fluidsimfoam.foam_input_files.ast import (
    Code,
    CodeStream,
    Dict,
    DimensionSet,
    FoamInputFile,
    List,
    Value,
)

DEFAULT_CODE_INCLUDE = '#include "fvCFD.H"'
DEFAULT_CODE_OPTIONS = (
    "-I$(LIB_SRC)/finiteVolume/lnInclude \\\n-I$(LIB_SRC)/meshTools/lnInclude"
)
DEFAULT_CODE_LIBS = "-lmeshTools \\\n-lfiniteVolume"


class FieldABC(ABC):
    cls: str

    @classmethod
    def from_code(cls, code: str, skip_boundary_field=False):
        if "nonuniform" not in code:
            tree = parse(code)
            return cls(None, None, tree=tree)

        index_nonuniform = code.index("nonuniform")
        index_boundaryField = code.rindex("boundaryField", index_nonuniform)
        index_opening_par = code.index("(", index_nonuniform, index_boundaryField)
        index_closing_par = code.rindex(
            ")", index_nonuniform, index_boundaryField
        )

        code_to_parse = code[:index_nonuniform] + ";\n"
        if not skip_boundary_field:
            code_to_parse += "\n" + code[index_boundaryField:]

        tree = parse(code_to_parse)
        code_data = code[index_opening_par + 1 : index_closing_par].strip()

        if code_data.startswith("("):
            code_data = re.sub("[()]", "", code_data)

        data = np.loadtxt(StringIO(code_data))

        tree.data = data
        return cls("", "", tree=tree, values=data)

    @classmethod
    def from_path(cls, path: str or Path, skip_boundary_field=False):
        path = Path(path)
        field = cls.from_code(
            path.read_text(), skip_boundary_field=skip_boundary_field
        )
        field.path = path
        return field

    def __init__(self, name, dimension, tree=None, values=None):
        if tree is not None:
            self.tree = tree
        else:
            info = {
                "version": "2.0",
                "format": "ascii",
                "class": self.cls,
                "object": name,
            }

            if not isinstance(dimension, DimensionSet):
                dimension = DimensionSet(dimension)

            self.tree = FoamInputFile(
                info, children={"dimensions": dimension, "internalField": None}
            )

            self.tree.set_child("boundaryField", {})

        if values is not None:
            self.set_values(values)

        self.path = None

    def dump(self):
        return self.tree.dump()

    def overwrite(self):
        if self.path is None:
            raise ValueError("self.path is None")
        with open(self.path, "w") as file:
            file.write(self.dump())

    @abstractmethod
    def set_values(self, values):
        """Set internalField with value(s)"""

    def set_codestream(
        self,
        code,
        include=DEFAULT_CODE_INCLUDE,
        options=DEFAULT_CODE_OPTIONS,
        libs=DEFAULT_CODE_LIBS,
    ):
        data = {
            "codeInclude": include,
            "codeOptions": options,
            "codeLibs": libs,
            "code": code,
        }
        data = {key: Code(key, value.strip()) for key, value in data.items()}
        self.tree.children["internalField"] = CodeStream(
            data, name="internalField", directive="#codeStream"
        )

    def set_boundary(self, name, type_, value=None, gradient=None):
        boundaries = self.tree.children["boundaryField"]
        data = {"type": type_}
        if value is not None:
            data["value"] = value
        if gradient is not None:
            data["gradient"] = gradient
        boundaries[name] = Dict(data, name=name)

    def set_name(self, name):
        self.tree.info["object"] = name

    def get_array(self):
        return np.array(self.tree.children["internalField"])


class VolScalarField(FieldABC):
    cls = "volScalarField"

    def set_values(self, values):
        if isinstance(values, Number):
            value = Value(values, name="uniform")
        else:
            value = List(
                list(values),
                name=f"internalField nonuniform\nList<scalar>\n{len(values)}",
            )
        self.tree.children["internalField"] = value


class VolVectorField(FieldABC):
    cls = "volVectorField"

    def set_values(self, values, vy=None, vz=None):
        if vy is not None:
            if vz is None:
                raise ValueError
            if not isinstance(values, np.ndarray):
                raise ValueError
            if values.ndim != 1 or values.size != vy.size != vz.size:
                raise ValueError
            vx = values
            values = np.stack([vx, vy, vz]).T

        n_elem = len(values)
        if n_elem == 3 and isinstance(values[0], Number):
            value = Value(
                "(" + " ".join(str(value) for value in values) + ")",
                name="uniform",
            )
        else:
            value = values
        self.tree.set_child("internalField", value)

    def get_components(self):
        arr = self.get_array()
        return arr[:, 0], arr[:, 1], arr[:, 2]


class VolTensorField(FieldABC):
    cls = "volTensorField"

    def set_values(self, values):
        if not isinstance(values, np.ndarray) or values.ndim != 2:
            raise NotImplementedError(
                "not isinstance(values, np.ndarray) or values.ndim != 2"
            )

        self.tree.set_child("internalField", values)


classes = {
    cls.cls: cls for cls in (VolScalarField, VolVectorField, VolTensorField)
}


def read_field_file(path):
    cls = None
    with open(path, "r") as file:
        for line in file:
            line = line.strip()
            if line.startswith("class "):
                cls = line.split()[-1][:-1]
                break

    if cls is None:
        raise RuntimeError(f"no class found for file {path}")

    cls = classes[cls]

    return cls.from_path(path)


def _isplit(source, sep="\n"):
    sepsize = len(sep)
    start = 0
    while True:
        idx = source.find(sep, start)
        if idx == -1:
            yield source[start:]
            return
        yield source[start:idx]
        start = idx + sepsize


def create_field_from_code(code):
    for line in _isplit(code):
        line = line.strip()
        if line.startswith("class "):
            cls = line.split()[-1][:-1]
            break

    if cls is None:
        raise RuntimeError(f"no class found for this code: {code[:500] = }")

    cls = classes[cls]

    return cls.from_code(code)



"""Helper to create decomposeParDict files"""

from fluidsimfoam.foam_input_files import FileHelper, FoamInputFile

supported_methods = set(
    [
        "simple",
        "scotch",
        "hierarchical",
    ]
)

methods = set(
    [
        "manual",
        "simple",
        "hierarchical",
        "kahip",
        "metis",
        "scotch",
        "structured",
        "multiLevel",
    ]
)


def check_method(method):
    if method not in methods:
        raise ValueError(f"{method = } not in {methods}")

    if method not in supported_methods:
        raise NotImplementedError(f"{method = } not in {supported_methods = }")


class DecomposeParDictHelper(FileHelper):
    def __init__(
        self,
        nsubdoms=1,
        method="scotch",
    ):
        self.nsubdoms = nsubdoms
        check_method(method)
        self.method = method

    def complete_params(self, params):
        par_params = params._set_child(
            "parallel",
            attribs={
                "nsubdoms": 1,
                "method": self.method,
                "nsubdoms_xyz": None,
                "order": "xyz",
                "delta": 0.001,
            },
            doc="TODO",
        )

        par_params._set_child("scotch", attribs={"strategy": None})

    def make_tree(self, params):
        par_params = params.parallel

        nsubdoms = par_params.nsubdoms
        if nsubdoms == 1:
            return False

        method = par_params.method
        check_method(method)

        tree = FoamInputFile(
            info={
                "version": 2.0,
                "format": "ascii",
                "class": "dictionary",
                "location": '"system"',
                "object": "decomposeParDict",
            }
        )

        data = {"numberOfSubdomains": nsubdoms, "method": method}

        key_coeffs_dict = "coeffs"
        if method == "simple":
            coeffs = {
                "n": par_params.nsubdoms_xyz,
                "order": par_params.order,
                "delta": par_params.delta,
            }
        elif method == "scotch":
            if par_params.scotch.strategy is None:
                coeffs = None
            else:
                coeffs = {"strategy": par_params.scotch.strategy}
        elif method == "hierarchical":
            coeffs = {"n": par_params.nsubdoms_xyz}
        else:
            raise RuntimeError

        if coeffs is not None:
            data[key_coeffs_dict] = coeffs

        tree.init_from_py_objects(data)

        return tree



"""Base class for the ``params`` object"""

import re

import fluiddyn.util.xmltotext
from fluidsim_core.params import Parameters as _Parameters


def get_indent_after_tag(text):
    return " " * (len(re.match(r"\s*\S*", text).group()) + 1)


def get_position_first_letter(text):
    return len(re.match(r"\s*", text).group())


pattern_entries = re.compile(r'''[\w]+="[^"]*"''')


def format_too_long_tagstart(text, lengthmax=79):
    if len(text) <= lengthmax:
        return text

    lines = []
    first_word = text.strip().split(None, 1)[0]

    position_first_letter = get_position_first_letter(text)
    line = " " * position_first_letter + first_word
    indent = get_indent_after_tag(text)

    entries = pattern_entries.findall(text)
    for word in entries:
        if len(line + word) > lengthmax:
            lines.append(line)
            line = indent + word
        else:
            if line.endswith(" "):
                line = line + word
            else:
                line = line + " " + word

    if text.endswith("/>"):
        end = "/>"
    else:
        end = ">"

    lines.append(line + end)
    return "\n".join(lines)


fluiddyn.util.xmltotext.format_too_long_tagstart = format_too_long_tagstart


table = {
    "(": "__opar__",
    ")": "__cpar__",
    ",": "__comma__",
    "*": "__mul__",
    ".": "__dot__",
    "|": "__bar__",
}
pattern_protect = re.compile(
    "(" + "|".join(re.escape(c) for c in table.keys()) + ")"
)
pattern_unprotect = re.compile("(" + "|".join(table.values()) + ")")
table_inverse = {v: k for k, v in table.items()}


def protect_characters(name):
    return pattern_protect.sub(lambda match: table[match.group()], name)


def unprotect_characters(name):
    return pattern_unprotect.sub(lambda match: table_inverse[match.group()], name)


class Parameters(_Parameters):
    def _make_element_xml(self, parentxml=None):
        elemxml = super()._make_element_xml(parentxml)
        elemxml.attrib = {
            protect_characters(name): value
            for name, value in elemxml.attrib.items()
        }
        return elemxml

    def _load_from_elemxml(self, elemxml):
        elemxml.attrib = {
            unprotect_characters(name): value
            for name, value in elemxml.attrib.items()
        }
        super()._load_from_elemxml(elemxml)

    def __repr__(self):
        return unprotect_characters(super().__repr__())

    def _del_attrib(self, key):
        del self.__dict__[key]
        self._key_attribs.remove(key)

    def __getitem__(self, key):
        try:
            return self.__getattribute__(key)
        except AttributeError:
            return super().__getitem__(key)

    def _update_attribs(self, attribs):
        for key, value in attribs.items():
            try:
                self[key] = value
            except AttributeError:
                self._set_attrib(key, value)



"""Read information on the mesh from the polyMesh directory

"""

import re
from functools import lru_cache
from io import StringIO

import numpy as np


@lru_cache(maxsize=1)
def get_points_coords(path):
    """Get points coordinates"""

    with open(path) as file:
        for line in file:
            if line == "FoamFile\n":
                break
        for line in file:
            if line == "}\n":
                break
        for line in file:
            line = line[:-1]
            if line.isdigit():
                nb_cells = int(line)
                break
        assert file.readline() == "(\n"
        txt = file.read()

    if txt.startswith("("):
        txt = re.sub("[()]", "", txt)

    txt = txt.strip()

    index_last_comment = txt.rfind("\n//")
    txt = txt[:index_last_comment].strip()

    coords = np.loadtxt(StringIO(txt))
    x = coords[:, 0]
    y = coords[:, 1]
    z = coords[:, 2]

    assert x.size == nb_cells

    return x, y, z



"""Internal machinery to generate the OpenFOAM input files"""

from abc import ABC, abstractmethod
from inspect import getmodule
from pathlib import Path

import jinja2
from inflection import camelize, underscore

from fluiddyn.util import import_class
from fluidsimfoam.foam_input_files import FileHelper, parse, read_field_file


class InputFiles:
    """Container of the generator objects"""

    def __init__(self, output):
        self.output = output
        mod = getmodule(type(output.sim))
        self.templates_dir = Path(mod.__file__).absolute().parent / "templates"

        loader = jinja2.ChoiceLoader(
            [
                jinja2.PackageLoader(mod.__package__, "templates"),
                jinja2.PackageLoader("fluidsimfoam", "resources"),
            ]
        )

        self.jinja_env = jinja2.Environment(
            loader=loader,
            undefined=jinja2.StrictUndefined,
            keep_trailing_newline=True,
        )

    def get_template(self, template_name):
        return self.jinja_env.get_template(template_name)


class FileGeneratorABC(ABC):
    rel_path: str

    def __init__(self, output):
        self.output = output
        self.input_files = output.input_files

    def generate_file(self, params=None):
        """Generate the file"""
        if params is None:
            params = self.output.sim.params

        code = self.generate_code(params)
        if code is False:
            return

        with open(self.output.path_run / self.rel_path, "w") as file:
            file.write(code)

    @abstractmethod
    def generate_code(self):
        """Generate the code of the file"""

    def read(self):
        path = self.output.path_run / self.rel_path
        if any(
            self.rel_path.startswith(start) for start in ["system", "constant"]
        ):
            tree = parse(path.read_text())
            tree.path = path
            return tree
        else:
            return read_field_file(path)

    def overwrite(self, dumpable):
        with open(self.output.path_run / self.rel_path, "w") as file:
            file.write(dumpable.dump())


class FileGenerator(FileGeneratorABC):
    template_name: str

    def __init__(self, output):
        super().__init__(output)

    def generate_code(self, params=None):
        """Generate the code of the file from ...

        - a method named like `sim.output._make_code_block_mesh_dict`,
        - or a Jinja template.
        """
        if params is None:
            params = self.output.sim.params

        try:
            make_code = getattr(self.output, "_make_code_" + self._name)
        except AttributeError:
            make_code = None

        if make_code is None:
            try:
                make_tree = getattr(self.output, "_make_tree_" + self._name)
            except AttributeError:
                make_tree = None

            if make_tree is None:
                try:
                    helper = getattr(self.output, "_helper_" + self._name)
                except AttributeError:
                    pass
                else:
                    if isinstance(helper, FileHelper):
                        make_tree = helper.make_tree

            if make_tree is not None:

                def make_code(params_):
                    tree = make_tree(params_)
                    if tree is False:
                        return False
                    return tree.dump()

        if make_code is None:
            template = self.input_files.get_template(self.template_name)
            return template.render(params=params)

        if (self.input_files.templates_dir / self.template_name).exists():
            raise RuntimeError(
                "Fluidsimfoam solver issue: "
                f"2 concurrent methods to produce {self.rel_path}:\n"
                f"- template in {self.input_files.templates_dir},\n"
                f"- function output._make_code_{self._name}.\n"
                "Remove the file or the function (or make it equal to None)."
            )

        return make_code(params)

    @classmethod
    def _complete_params_with_default(cls, params, info_solver):
        output_cls = import_class(
            info_solver.classes.Output.module_name,
            info_solver.classes.Output.class_name,
        )
        try:
            complete_params = getattr(output_cls, "_complete_params_" + cls._name)
        except AttributeError:
            complete_params = None

        if complete_params is None:
            try:
                helper = getattr(output_cls, "_helper_" + cls._name)
            except AttributeError:
                pass
            else:
                if isinstance(helper, FileHelper):
                    complete_params = helper.complete_params

        if complete_params is not None:
            complete_params(params)


def new_file_generator_class(file_name, dir_name="0"):
    cls_name = f"FileGenerator{camelize(file_name)}"
    return type(
        cls_name,
        (FileGenerator,),
        {
            "dir_name": dir_name,
            "rel_path": f"{dir_name}/{file_name}",
            "template_name": f"{file_name}.jinja",
            "_name": underscore(file_name.replace(".", "_")),
        },
    )



"""Testing utilities"""

import shutil
from pathlib import Path

import pytest

from fluidsimfoam.foam_input_files import dump, parse


class skipif_executable_not_available:
    def __init__(self, command_name):
        path_foam_executable = shutil.which(command_name)

        self.skipif = pytest.mark.skipif(
            path_foam_executable is None,
            reason=f"executable '{command_name}' not available",
        )

    def __call__(self, func):
        return self.skipif(func)


def check_saved_case(path_saved_case, path_run, files_compare_tree=None):
    if files_compare_tree is None:
        files_compare_tree = []

    paths_in_sim = [
        path.relative_to(path_saved_case)
        for path in path_saved_case.rglob("*")
        if not path.is_dir()
        and not any(path.name.startswith(s) for s in ("README", "All"))
        and not path.parent.name == "polyMesh"
    ]

    for relative_path in paths_in_sim:
        path_manual = path_saved_case / relative_path
        text_manual = path_manual.read_text()
        if str(relative_path.parent) == "0_org":
            relative_path = Path("0") / relative_path.name
        path_produced = path_run / relative_path
        assert path_produced.exists(), relative_path
        text_produced = path_produced.read_text()

        if relative_path.name in files_compare_tree:
            tree_saved_file = parse(text_manual)
            tree_from_py = parse(text_produced)
            assert dump(tree_saved_file).strip() == dump(tree_from_py).strip()
        else:
            assert text_produced == text_manual, relative_path



"""Abstract Syntax Trees for OpenFOAM input files"""

import re
from abc import ABC, abstractmethod
from dataclasses import dataclass
from numbers import Number
from textwrap import dedent
from typing import Optional

import numpy as np
from inflection import underscore

symbols = ["kg", "m", "s", "K", "kmol", "A", "cd"]


def str2foam_units(units):
    result = [0] * 7
    sign = 1
    while units:
        try:
            index = re.search(r"[/.]", units).start()
        except AttributeError:
            unit_all = units
            units = ""
            next_oper = "."
        else:
            next_oper = units[index]
            assert next_oper in "/."
            unit_all = units[:index]
            units = units[index + 1 :]
        if "^" in unit_all:
            unit_name, unit_value = unit_all.split("^")
            unit_value = int(unit_value)
        else:
            unit_name, unit_value = unit_all, 1
        if unit_name != "1":
            unit_index = symbols.index(unit_name)
            result[unit_index] = sign * unit_value
        sign = 1 if next_oper == "." else -1
    return result


def foam_units2str(foam_units):
    result = []
    for symbol, exponent in zip(symbols, foam_units):
        if exponent == 0:
            continue
        operator = "." if exponent > 0 else "/"
        if abs(exponent) == 1:
            result.append(f"{operator}{symbol}")
        else:
            result.append(f"{operator}{symbol}^{abs(exponent)}")
    result = "".join(result)
    if result.startswith("."):
        result = result[1:]
    elif result.startswith("/"):
        result = "1" + result
    return result


def _compute_spaces_to_align(data, max_length=20):
    try:
        max_length = min(
            max_length,
            max(
                len(key)
                for key, value in data.items()
                if value is not None and not isinstance(value, (Dict, List))
            ),
        )
    except ValueError:
        max_length = 0

    default_space = 4
    return max_length + default_space


class Node:
    def __eq__(self, other):
        if type(other) is type(self):
            return self.__dict__ == other.__dict__
        return False


class NodeLikePyDict(ABC):
    def init_from_py_objects(
        self,
        data: dict,
        dimensions: Optional[dict] = None,
        default_dimension=False,
        comments: Optional[dict] = None,
    ):
        if dimensions is None:
            dimensions = {}
        if comments is None:
            comments = {}
        for key, value in data.items():
            if isinstance(value, (type(None), str)):
                self.set_child(key, value)
            elif isinstance(value, (Number, list)):
                if default_dimension is False:
                    self.set_child(key, value)
                else:
                    dimension = dimensions.get(key, default_dimension)
                    if isinstance(value, list):
                        value = List(value)
                    self.set_value(key, value, dimension)
            elif isinstance(value, dict):
                dimensions_dict = dimensions.get(key, None)
                comments_dict = comments.get(key, None)
                obj = Dict({}, name=key, comments=comments_dict)
                obj.init_from_py_objects(
                    value,
                    dimensions=dimensions_dict,
                    default_dimension=default_dimension,
                    comments=comments_dict,
                )
                self._set_item(key, obj)
            else:
                raise NotImplementedError(type(value))

    def set_child(self, key, child):
        if (
            isinstance(child, list)
            and child
            and isinstance(child[0], list)
            and child[0]
            and all(isinstance(n, Number) for n in child[0])
        ):
            child = np.array(child)

        if isinstance(child, (type(None), str, Number, DimensionSet)):
            pass
        elif isinstance(child, dict):
            child = Dict(child, name=key)
        elif isinstance(child, list):
            child = List(child, name=key)
        elif isinstance(child, np.ndarray):
            shape = child.shape
            ndim = child.ndim
            child.tolist()
            dtype = None
            if ndim == 2:
                if shape[1] == 9:
                    dtype = "tensor"
                elif shape[1] == 3:
                    dtype = "vector"
                else:
                    raise NotImplementedError
                child = [List(sequence) for sequence in child]
            elif ndim == 1:
                dtype = "scalar"
            else:
                raise NotImplementedError
            child = List(child, name=key, dtype=dtype)
        elif isinstance(child, Node):
            pass
        else:
            raise NotImplementedError(type(child))
        self._set_item(key, child)

    def set_value(self, name, value, dimension=None):
        if isinstance(value, Number) or dimension is not None:
            value = Value(value, name, dimension=dimension)
        self._set_item(name, value)

    @abstractmethod
    def _set_item(self, key, value):
        """Set an item"""


class FoamInputFile(Node, NodeLikePyDict):
    def __init__(self, info, children=None, header=None, comments=None):
        self.info = info
        if children is None:
            children = {}
        self.children = children
        self.header = header
        self.comments = comments
        self.path = None

    def __repr__(self):
        tmp = ["InputFile(\n"]
        if self.info is not None:
            tmp.append(f"info={self.info},\n")
        tmp.append(f"children={self.children}\n)")
        return "".join(tmp)

    def dump(self):
        tmp = []
        if self.info is not None:
            tmp1 = ["FoamFile\n{"]
            for key, node in self.info.items():
                s = (12 - len(key)) * " "
                tmp1.append(f"    {key}{s}{node};")
            tmp1.append("}")
            tmp.append("\n".join(tmp1))

        num_spaces = _compute_spaces_to_align(self.children, max_length=14)
        for key, node in self.children.items():
            if hasattr(node, "dump"):
                code_node = node.dump()
                # special for isolated list
                if key is None and isinstance(node, List):
                    code_node += ";"
            elif node is None:
                code_node = f"{key}"
            else:
                if hasattr(node, "dump_without_assignment"):
                    node_dumped = node.dump_without_assignment()
                else:
                    node_dumped = node
                if node_dumped == "":
                    s = ""
                else:
                    s = max(2, (num_spaces - len(key))) * " "
                code_node = f"{key}{s}{node_dumped};"
            if self.comments is not None and key in self.comments:
                comment = self.comments[key]
                if isinstance(comment, str):
                    comment = "// " + comment.replace("\n", "\n// ")
                    code_node = comment + "\n" + code_node
            tmp.append(code_node)
        result = "\n\n".join(tmp)
        if self.header is not None:
            result = self.header + "\n" + result
        if result[-1] != "\n":
            result += "\n"
        return result

    def overwrite(self):
        if self.path is None:
            raise ValueError("self.path is None")
        with open(self.path, "w") as file:
            file.write(self.dump())

    def _set_item(self, key, value):
        self.children[key] = value

    def __setitem__(self, key, item):
        self.children[key] = item

    def __getitem__(self, key):
        return self.children[key]


@dataclass
class Assignment:
    name: str
    value: object

    def dump(self, indent=0):
        if hasattr(self.value, "dump"):
            return self.value.dump(indent)
        else:
            return indent * " " + f"{self.name}  {self.value};"


class VariableAssignment(Assignment):
    def __repr__(self):
        s = super().__repr__()
        return s

    def dump(self, indent=0):
        if hasattr(self.value, "dump"):
            return indent * " " + f"{self.name}  {self.value.dump(indent)};"
        else:
            return indent * " " + f"{self.name}  {self.value};"


@dataclass
class Value(Node):
    def __init__(self, value, name=None, dimension=None):
        self.value = value
        self.name = name
        if isinstance(dimension, (list, tuple)):
            dimension = foam_units2str(dimension)
        self.dimension = dimension

    def __repr__(self):
        if self.dimension is not None and self.name is not None:
            return f'Value({self.value}, name="{self.name}", dimension="{self.dimension}")'
        elif self.dimension is None and self.name is not None:
            return f'Value({self.value}, name="{self.name}")'
        elif self.dimension is not None and self.name is None:
            return f'Value({self.value}, dimension="{self.dimension}")'
        else:
            return f"Value({self.value})"

    def dump_without_assignment(self, indent=0):
        if self.dimension is not None:
            dimension_list = str2foam_units(self.dimension)
            dimension_dumped = " ".join(str(number) for number in dimension_list)

        if isinstance(self.value, Node) and hasattr(self.value, "dump"):
            value_dumped = self.value.dump()
        else:
            value_dumped = str(self.value)

        if self.dimension is not None and self.name is not None:
            return f"{self.name} [{dimension_dumped}] {value_dumped}"
        elif self.dimension is None and self.name is not None:
            return f"{self.name} {value_dumped}"
        elif self.dimension is not None and self.name is None:
            return f"[{dimension_dumped}] {value_dumped}"
        else:
            return f"{value_dumped}"


class DimensionSet(list, Node):
    def __init__(self, foam_units):
        if isinstance(foam_units, str):
            foam_units = str2foam_units(foam_units)
        if not all(isinstance(elem, int) for elem in foam_units):
            raise ValueError("Bad {foam_units = }")
        super().__init__(foam_units)

    def __repr__(self):
        return foam_units2str(self)

    def dump_without_assignment(self, indent=0):
        return "[" + " ".join(str(number) for number in self) + "]"


class Dict(dict, Node, NodeLikePyDict):
    def __init__(self, data, name=None, directive=None, comments=None):
        self._name = name
        self._directive = directive
        self.comments = comments
        super().__init__(**data)

    def get_name(self):
        return self._name

    def __repr__(self):
        return super().__repr__()

    def dump(self, indent=0):
        tmp = []
        indentation = indent * " "
        if self._name is not None:
            line = indentation + self._name
            if self._directive is not None:
                line += "  " + self._directive
            tmp.append(line + f"\n{indentation}" + "{")

        num_spaces = _compute_spaces_to_align(self)
        for key, node in self.items():
            if self.comments is not None and key in self.comments:
                comment = self.comments[key]
                if isinstance(comment, str):
                    tmp.append("    // " + comment.replace("\n", "\n    // "))

            if hasattr(node, "dump"):
                tmp.append(node.dump(indent + 4))
            elif node is None:
                tmp.append(indentation + f"    {key}")
            else:
                if hasattr(node, "dump_without_assignment"):
                    code_node = node.dump_without_assignment()
                else:
                    code_node = node
                if node == "":
                    s = ""
                else:
                    s = max(2, (num_spaces - len(key))) * " "
                tmp.append(indentation + f"    {key}{s}{code_node};")
        tmp.append(indentation + "}")

        # because OpenFOAM inconsistency
        if isinstance(self, CodeStream):
            tmp[-1] += ";"

        return "\n".join(tmp)

    def _set_item(self, key, value):
        self[key] = value


class List(list, Node):
    """Represents an OpenFoam list"""

    def __init__(self, iterable=None, name=None, dtype=None):
        self._name = name
        self._dtype = dtype
        super().__init__(iterable)

    def get_name(self):
        return self._name

    def add_name(self, name):
        if self._name is None:
            self._name = name
        elif isinstance(self._name, str):
            if self._name != name:
                self._name = name + " " + self._name
        else:
            raise RuntimeError()

    def __repr__(self):
        return super().__repr__()

    def _make_list_strings(self, indent):
        return [self._dump_item(item, indent) for item in self]

    def _dump_item(self, item, indent=0):
        if isinstance(item, (Node, Assignment)) and hasattr(item, "dump"):
            return item.dump(indent)
        else:
            return indent * " " + str(item)

    def dump(self, indent=0):
        tmp = []
        indentation = indent * " "
        if self._name is None:
            tmp.extend(self._make_list_strings(indent=0))
            return indentation + "(" + " ".join(tmp) + ")"
        else:
            header = self._name
            if self._dtype is not None:
                header += (
                    f"   nonuniform List<{self._dtype}>\n"
                    f"{indentation}{len(self)}"
                )

            tmp.append(indentation + header + f"\n{indentation}" + "(")
            special_keyss = {
                "blocks": ("hex",),
                "edges": ("spline", "arc", "polyLine", "BSpline", "line"),
            }
            if self._name not in special_keyss.keys():
                tmp.append("\n".join(self._make_list_strings(indent + 4)))
            elif self:
                special_keys = special_keyss[self._name]
                if not self[0] in special_keys:
                    raise ValueError(self)
                special_key = self[0]
                lines = []
                items_line = None
                for item in self:
                    if item == special_key:
                        if items_line is not None:
                            lines.append(items_line)
                        items_line = [item]
                    else:
                        items_line.append(item)
                lines.append(items_line)
                lines = [
                    " ".join(self._dump_item(_item) for _item in items_line)
                    for items_line in lines
                ]
                tmp.extend((indent + 4) * " " + line for line in lines)
            tmp.append(indentation + ");")
            return "\n".join(tmp)


class CodeStream(Dict):
    """A dictionnary to store #codeStream"""


def _make_alias(name):
    def get(self):
        return self[name].code

    def set(_self, value):
        _self[name].code = value

    return property(get, set)


for _name in ("codeInclude", "codeOptions", "codeLibs", "code"):
    setattr(CodeStream, underscore(_name), _make_alias(_name))


class Code(Node):
    def __init__(self, name, code, directive=None):
        self.name = name
        self.code = dedent(code)
        self.directive = directive

    def __repr__(self):
        if self.directive is None:
            return f'Code("{self.code[:20]}[...]")'
        return f'Code("{self.code[:20]}[...]", directive={self.directive})'

    def dump(self, indent=0):
        tmp = []
        indentation = indent * " "
        indentation4 = (indent + 4) * " "
        start = indentation + self.name
        if self.directive is not None:
            start += " " + self.directive
        tmp.append(start + f"\n{indentation}" + "#{")
        for line in self.code.split("\n"):
            tmp.append(indentation4 + line)
        tmp.append(indentation + "#};")
        return "\n".join(tmp)


@dataclass
class Name(Node):
    def __init__(self, name, value=None):
        self.name = name
        self.value = value

    def __repr__(self):
        return f"Name({self.name})"

    def dump(self, indent=0):
        return f"{self.name}"


class Directive(Node):
    def __init__(self, directive, content):
        self.directive = directive
        self.content = content

    def __repr__(self):
        return f"{self.directive}  {self.content}"

    def dump(self, indent=0):
        return indent * " " + f"{self.directive}  {self.content};"



"""Format OpenFOAM input files"""

import lark

from fluidsimfoam.foam_input_files.fields import create_field_from_code
from fluidsimfoam.foam_input_files.parser import dump, parse


def _unidiff_output(expected, actual):
    """
    Returns a string containing the unified diff of two multiline strings.

    Taken from https://stackoverflow.com/a/845432
    """
    import difflib

    expected = expected.splitlines(keepends=True)
    actual = actual.splitlines(keepends=True)
    diff = difflib.unified_diff(expected, actual)
    return "".join(diff)


class FoamFormatError(RuntimeError):
    pass


def format_code(code, as_field=False, check=True):
    if not as_field:
        tree = parse(code)
        result = dump(tree)
        if check:
            try:
                tree1 = parse(result)
            except lark.exceptions.LarkError:
                raise FoamFormatError(f"LarkError while formatting code\n{code}")
            else:
                result1 = dump(tree1)

            if result != result1:
                print(
                    "Not able to correctly parse this code:\n"
                    + _unidiff_output(result, result1)
                )
                raise FoamFormatError("Not able to correctly parse this code")
        return result
    else:
        return create_field_from_code(code).dump()



"""Class for the ``sim.output.log`` object"""

import re

import matplotlib.pyplot as plt
import numpy as np

from fluidsim_core.output.remaining_clock_time import RemainingClockTime


def get_log_tail(path_file, nbytes=1000):
    log_size = path_file.stat().st_size
    with open(path_file, "r") as file:
        file.seek(max(0, log_size - nbytes))
        return file.read()


def read_time_last(path_file):
    text = get_log_tail(path_file, 1000)
    index = text.rfind("\nTime = ")
    if index == -1:
        print("'Time = ' not found in log file")
        return None
    text = text[index + 8 :]
    return float(text.split(None, 1)[0])


class Log(RemainingClockTime):
    _tag = "log"

    def __init__(self, output):
        self.output = output
        self._path_file = None

    def _load_times(self):
        """Load remaining time data.

        - equation_times
        - remaining_clock_times
        - clock_times_per_timestep
        - equation_time_start
        - full_clock_time

        """
        text = self.text

        match = re.search(r"start_time = ([\d\.]+)\n", text)
        if match is not None:
            equation_time_start = float(match.groups()[0])
        else:
            equation_time_start = 0.0

        match = re.search(r"end_time = ([\d\.]+)\n", text)
        if match is not None:
            eq_time_end = float(match.groups()[0])
        else:
            raise RuntimeError

        eq_times = re.findall(r"\nTime = ([\d\.]+e?[-\d]*)", text)
        clock_times = re.findall(r"\nExecutionTime = ([\d\.]+)", text)
        eq_times = eq_times[: len(clock_times)]

        eq_times = np.array([float(word) for word in eq_times])
        clock_times = np.array([float(word) for word in clock_times])

        estimation_clock_time_per_time_step = clock_times[-1] / len(clock_times)

        indices_time_step = np.arange(clock_times.size)

        # remove times with same clock time
        clock_times, indices_unique = np.unique(clock_times, return_index=True)
        eq_times = eq_times[indices_unique]
        indices_time_step = indices_time_step[indices_unique]

        # decimate if needed
        precision_clock = 0.01
        step = round(
            min(8 * precision_clock, (clock_times.max() - clock_times.min()) / 2)
            / estimation_clock_time_per_time_step
        )
        if step > 1:
            clock_times = clock_times[::step]
            eq_times = eq_times[::step]
            indices_time_step = indices_time_step[::step]

        delta_eq_times = np.diff(eq_times)

        eq_times = eq_times[:-1]
        remaining_eq_times = eq_time_end - eq_times

        delta_clock_times = np.diff(clock_times)

        remaining_clock_times = (
            remaining_eq_times / delta_eq_times * delta_clock_times
        )

        data = {
            "equation_times": eq_times,
            "remaining_clock_times": remaining_clock_times,
            "clock_times_per_timestep": delta_clock_times
            / np.diff(indices_time_step),
            "equation_time_start": equation_time_start,
            "full_clock_time": clock_times[-1],
            "remaining_eq_times": remaining_eq_times,
            "delta_eq_times": delta_eq_times,
            "delta_clock_times": delta_clock_times,
        }

        return data

    @property
    def path_file(self):
        output = self.output
        if output and not self._path_file:
            path_run = output.path_run
            logfiles = sorted(path_run.glob("log*.txt"))
            if logfiles:
                self._path_file = logfiles[-1]
            else:
                print("No log file found")
                return None

        return self._path_file.resolve()

    @path_file.setter
    def path_file(self, path_log_file):
        self._path_file = path_log_file

    @property
    def text(self):
        if self.path_file is None:
            return None

        with open(self.path_file) as file:
            return file.read()

    def get_log_tail(self, nbytes=1000):
        if self.path_file is None:
            raise IOError(f"No log file found in {self.output.path_run}")
        return get_log_tail(self.path_file, nbytes)

    @property
    def time_last(self):
        if self.path_file is None:
            return None
        return read_time_last(self.path_file)

    def _choose_variable_name(self, variable_name):
        if (
            variable_name is not None
            and variable_name not in self.output.name_variables
        ):
            raise ValueError(
                f"variable_name has to be in {self.output.name_variables}"
            )

        if variable_name is None:
            for key in ("p", "p_rbgh"):
                if key in self.output.name_variables:
                    variable_name = key
                    break
        return variable_name

    def get_last_residual(self, variable_name=None):
        if self.path_file is None:
            return None
        text = self.get_log_tail(5000)

        variable_name = self._choose_variable_name(variable_name)
        matches = list(
            re.finditer(
                r"\nTime = (?P<time>[\d\.]+e?-?[\d]*)"
                rf"[\s\S]+?Solving for {variable_name}, "
                r"Initial residual = (?P<initial>[\d\.]+e?-?[\d]*)",
                text,
            )
        )

        if not matches:
            raise ValueError(f"No match found for {variable_name}, text=\n{text}")

        match = matches[-1]
        data = match.groupdict()
        time = float(data["time"])
        residual = float(data["initial"])
        return time, residual

    def plot_residuals(self, variable_name=None, tmin=0.0):
        variable_name = self._choose_variable_name(variable_name)
        matches = list(
            re.finditer(
                r"\nTime = (?P<time>[\d\.]+e?-?[\d]*)"
                rf"[\s\S]+?Solving for {variable_name}, "
                r"Initial residual = (?P<initial>[\d\.]+e?-?[\d]*)",
                self.text,
            )
        )

        times = np.empty(len(matches))
        residuals = np.empty_like(times)

        for index, match in enumerate(matches):
            data = match.groupdict()

            times[index] = data["time"]
            residuals[index] = data["initial"]

        fig, ax = plt.subplots()

        residuals = residuals[times > tmin]
        times = times[times > tmin]
        ax.plot(times, residuals)
        ax.set_xlabel("equation time")
        fig.suptitle(f"Initial residuals {variable_name}")
        return times, residuals



"""Helper to create blockMeshDict files

Taken from https://github.com/takaakiaoki/ofblockmeshdicthelper (Git commit
58589f1) and modified as follow:

- ``pyupgrade __init__.py --py39-plus``
- tested (100% coverage) and reproducible (sets sorted)
- refactor
- support for cyclic boundaries

"""

from itertools import groupby
from string import Template

from .. import DEFAULT_HEADER
from .edges import ArcEdge, SplineEdge
from .grading import EdgeGrading, Grading, SimpleGrading, SimpleGradingElement

__all__ = [
    "Grading",
    "SimpleGradingElement",
    "SimpleGrading",
    "EdgeGrading",
    "ArcEdge",
    "SplineEdge",
]


class Vertex:
    def __init__(self, x, y, z, name, index=None):
        self.x = x
        self.y = y
        self.z = z
        self.name = name  # identical name
        self.alias = {name}  # aliasname, self.name should be included

        # sequential index which is assigned at final output
        # for blocks, edges, boundaries
        self.index = index

    def format(self):
        comment = f"{self.index} {self.name}"
        if len(self.alias) > 1:
            comment += " : " + " ".join(sorted(self.alias))
        return f"( {self.x: .15g} {self.y: .15g} {self.z: .15g} )  // {comment}"

    def __repr__(self):
        return f"Vertex({self.x}, {self.y}, {self.z}, {self.name})"

    def copy(self):
        return type(self)(self.x, self.y, self.z, self.name)

    def __lt__(self, rhs):
        return (self.z, self.y, self.x) < (rhs.z, rhs.y, rhs.x)

    def __eq__(self, rhs):
        return (self.z, self.y, self.x) == (rhs.z, rhs.y, rhs.x)

    def __hash__(self):
        return hash((self.z, self.y, self.x))


class Point:
    def __init__(self, x, y, z):
        self.x = x
        self.y = y
        self.z = z

    def format(self):
        return f"( {self.x:18.15g} {self.y:18.15g} {self.z:18.15g} )"


class Face:
    def __init__(self, vnames, name):
        """
        vname is list or tuple of vertex names
        """
        self.vnames = vnames
        self.name = name

    def format(self, vertices):
        """Format instance to dump
        vertices is dict of name to Vertex
        """
        index = " ".join(str(vertices[vn].index) for vn in self.vnames)
        comment = " ".join(self.vnames)
        return f"({index:s})  // {self.name:s} ({comment:s})"


class HexBlock:
    def __init__(self, vnames, cells, name, grading=SimpleGrading(1, 1, 1)):
        """Initialize HexBlock instance
        vnames is the vertex names in order descrived in
            http://www.openfoam.org/docs/user/mesh-description.php
        cells is number of cells devied into in each direction
        name is the uniq name of the block
        grading is grading method.
        """
        self.vnames = vnames
        self.cells = cells
        self.name = name
        self.grading = grading

    def format(self, vertices):
        """Format instance to dump
        vertices is dict of name to Vertex
        """
        index = " ".join(str(vertices[vn].index) for vn in self.vnames)
        comment = " ".join(self.vnames)
        return (
            f"hex ({index}) {self.name} "
            f"({self.cells[0]:d} {self.cells[1]:d} {self.cells[2]:d}) "
            f"{self.grading.format()}  // {self.name} ({comment})"
        )

    def face(self, index, name=None):
        """Generate Face object
        index is number or keyword to identify the face of Hex
            0 = 'w' = 'xm' = '-100' = (0 4 7 3)
            1 = 'e' = 'xp' = '100' = (1 2 6 5)
            2 = 's' = 'ym' = '0-10' = (0 1 5 4)
            3 = 'n' = 'yp' = '010' = (2 3 7 6)
            4 = 'b' = 'zm' = '00-1' = (0 3 2 1) = "bottom"
            5 = 't' = 'zp' = '001' = (4 5 6 7) = "top"
        name is given to Face instance. If omitted, name is automatically
            genaratied like ('f-' + self.name + '-w')
        """
        kw_to_index = {
            "w": 0,
            "xm": 0,
            "-100": 0,
            "e": 1,
            "xp": 1,
            "100": 1,
            "s": 2,
            "ym": 2,
            "0-10": 2,
            "n": 3,
            "yp": 3,
            "010": 3,
            "b": 4,
            "bottom": 4,
            "zm": 4,
            "00-1": 4,
            "t": 5,
            "top": 5,
            "zp": 5,
            "001": 5,
        }
        index_to_vertex = [
            (0, 4, 7, 3),
            (1, 2, 6, 5),
            (0, 1, 5, 4),
            (2, 3, 7, 6),
            (0, 3, 2, 1),
            (4, 5, 6, 7),
        ]
        index_to_defaultsuffix = "wesnbt"

        if isinstance(index, str):
            index = kw_to_index[index]

        vnames = tuple([self.vnames[i] for i in index_to_vertex[index]])
        if name is None:
            parts = ["f"]
            if self.name:
                parts.append(self.name)
            parts.append(index_to_defaultsuffix[index])
            name = "-".join(parts)
        return Face(vnames, name)

    def faces(self, *args):
        return [self.face(arg) for arg in args]


class Boundary:
    def __init__(self, type_, name, faces=None, neighbour=None):
        """initialize boundary
        type_ is type keyword (wall, patch, empty, ..)
        name is nave of boundary emelment
        faces is faces which are applied with this boundary conditions
        """
        self.type_ = type_
        self.name = name
        if faces is None:
            faces = []
        elif isinstance(faces, Face):
            faces = [faces]
        self.faces = faces
        self.neighbour = neighbour

    def add_face(self, face):
        """add face instance
        face is a Face instance (not name) to be added
        """
        self.faces.append(face)

    def format(self, vertices):
        """Format instance to dump
        vertices is dict of name to Vertex
        """
        tmp = [self.name]
        if self.neighbour is None:
            tmp.append("{\n" + f"    type {self.type_};\n    faces\n    (")
        else:
            tmp.append(
                "{\n"
                + f"    type {self.type_};\n    neighbourPatch  {self.neighbour};\n    faces\n    ("
            )
        for face in self.faces:
            tmp.append(f"        {face.format(vertices)}")
        tmp.append("    );\n}")
        return "\n".join(tmp)


class BlockMeshDict:
    def __init__(self):
        self.convert_to_meters = 1.0
        self.vertices = {}  # mapping of uniq name to Vertex object
        self.blocks = {}
        self.edges = {}
        self.boundaries = {}
        self._vertices_in_blockmesh = None

    def set_metric(self, metric):
        """set self.comvert_to_meters by word"""
        metricsym_to_conversion = {
            "km": 1000,
            "m": 1,
            "cm": 0.01,
            "mm": 0.001,
            "um": 1e-6,
            "nm": 1e-9,
            "A": 1e-10,
            "Angstrom": 1e-10,
        }
        self.convert_to_meters = metricsym_to_conversion[metric]

    def set_scale(self, scale):
        self.convert_to_meters = scale

    def add_vertex(self, x, y=None, z=None, name=None):
        """add vertex by coordinate and uniq name
        x y z is coordinates of vertex
        name is uniq name to refer the vertex
        returns Vertex object whici is added.
        """
        if isinstance(x, Vertex):
            if any(arg is not None for arg in (y, z, name)):
                raise ValueError(
                    "x is a Vertex and "
                    "`any(arg is not None for arg in (y, z, name))`"
                )
            vertex = x
            name = vertex.name
        else:
            vertex = Vertex(x, y, z, name)
        self.vertices[name] = vertex
        return self.vertices[name]

    def del_vertex(self, name):
        """del name key from self.vertices"""
        del self.vertices[name]

    def reduce_vertex(self, name1, *names):
        """treat name1, name2, ... as same point.

        name2.alias, name3.alias, ... are merged with name1.alias
        the key name2, name3, ... in self.vertices are kept and mapped to
        same Vertex instance as name1
        """
        v = self.vertices[name1]
        for n in names:
            w = self.vertices[n]
            v.alias.update(w.alias)
            # replace mapping from n w by to v
            self.vertices[n] = v

    def merge_vertices(self):
        """call reduce_vertex on all vertices with identical values."""

        # groupby expects sorted data
        sorted_vertices = sorted(
            list(self.vertices.items()), key=lambda v: hash(v[1])
        )
        groups = []
        for k, g in groupby(sorted_vertices, lambda v: hash(v[1])):
            groups.append(list(g))
        for group in groups:
            if len(group) == 1:
                continue
            names = [v[0] for v in group]
            self.reduce_vertex(*names)

    def replicate_vertices_further_z(self, dz, add_to_name="_dz"):
        """Helper for 2d meshes"""
        for vertex_z0 in self.vertices.copy().values():
            vertex_z1 = vertex_z0.copy()
            vertex_z1.z = dz
            vertex_z1.name += add_to_name
            self.add_vertex(vertex_z1)

    def add_hexblock(
        self, vnames, cells, name=None, grading=SimpleGrading(1, 1, 1)
    ):
        if name is None:
            name = f"b{len(self.blocks)}"
        b = HexBlock(vnames, cells, name, grading)
        self.blocks[name] = b
        return b

    def add_arcedge(self, vnames, name, inter_vertex):
        e = ArcEdge(vnames, name, inter_vertex)
        self.edges[name] = e
        return e

    def add_splineedge(self, vnames, name, points):
        e = SplineEdge(vnames, name, points)
        self.edges[name] = e
        return e

    def add_boundary(self, type_, name, faces=None, neighbour=None):
        b = Boundary(type_, name, faces, neighbour)
        self.boundaries[name] = b
        return b

    def add_cyclic_boundaries(self, name0, name1, faces0, faces1):
        """Add 2 cyclic boundaries

        Example
        -------

        2 cyclic boundaries can be created as follow::

          add_cyclic_boundaries("outlet", "inlet", b0.face("e"), b0.face("w"))

        """
        b0 = self.add_boundary("cyclic", name0, faces0, neighbour=name1)
        b1 = self.add_boundary("cyclic", name1, faces1, neighbour=name0)
        return b0, b1

    def assign_vertexid(self, sort=True):
        """1. create list of Vertex which are referred by blocks only.
        2. sort vertex according to (x, y, z)
        3. assign sequence number for each Vertex
        4. sorted list is saved as self._vertices_in_blockmesh
        """
        # gather 'uniq' names which are referred by blocks
        vnames_kept = set()
        self._vertices_in_blockmesh = []
        for b in self.blocks.values():
            for n in b.vnames:
                v = self.vertices[n]
                if v.name not in vnames_kept:
                    vnames_kept.add(v.name)
                    self._vertices_in_blockmesh.append(v)
        if sort == "as_added":
            self._vertices_in_blockmesh = []
            for vname, v in self.vertices.items():
                if vname in vnames_kept:
                    self._vertices_in_blockmesh.append(v)
        elif sort:
            self._vertices_in_blockmesh = sorted(self._vertices_in_blockmesh)
        for i, v in enumerate(self._vertices_in_blockmesh):
            v.index = i

    def format_vertices_section(self):
        """format vertices section.
        assign_vertexid() should be called before this method, because
        self._vertices_in_blockmesh should be available and
        members of self._vertices_in_blockmesh should have valid index.
        """
        tmp = ["vertices\n("]
        for v in self._vertices_in_blockmesh:
            tmp.append("    " + v.format())
        tmp.append(");")
        return "\n".join(tmp)

    def format_blocks_section(self):
        """format blocks section.
        assign_vertexid() should be called before this method, because
        vertices refered by blocks should have valid index.
        """
        tmp = ["blocks\n("]
        for b in self.blocks.values():
            tmp.append("    " + b.format(self.vertices))
        tmp.append(");")
        return "\n".join(tmp)

    def format_edges_section(self):
        """format edges section.
        assign_vertexid() should be called before this method, because
        vertices refered by blocks should have valid index.
        """

        tmp = ["edges\n("]
        for e in self.edges.values():
            tmp.append("    " + e.format(self.vertices))
        tmp.append(");")
        return "\n".join(tmp)

    def format_boundary_section(self):
        """format boundary section.
        assign_vertexid() should be called before this method, because
        vertices refered by faces should have valid index.
        """

        tmp = ["boundary\n("]
        for b in self.boundaries.values():
            # format Boundary instance and add indent
            indent = " " * 4
            s = b.format(self.vertices).replace("\n", "\n" + indent)
            tmp.append(indent + s)
        tmp.append(");")
        return "\n".join(tmp)

    def format_mergepatchpairs_section(self):
        # not yet implemented
        return ""

    #         return """\
    # mergePatchPairs
    # (
    # );"""

    def format(self, header=DEFAULT_HEADER, sort_vortices=True):
        self.assign_vertexid(sort=sort_vortices)
        template = Template(
            r"""$header
FoamFile
{
    version     2.0;
    format      ascii;
    class       dictionary;
    object      blockMeshDict;
}
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

scale   $metricconvert;

$vertices

$blocks

$edges

$boundary
$mergepatchpairs
// ************************************************************************* //
"""
        )

        return template.substitute(
            header=header,
            metricconvert=str(self.convert_to_meters),
            vertices=self.format_vertices_section(),
            edges=self.format_edges_section(),
            blocks=self.format_blocks_section(),
            boundary=self.format_boundary_section(),
            mergepatchpairs=self.format_mergepatchpairs_section(),
        )


class BlockMeshDictRectilinear(BlockMeshDict):
    def __init__(self, lx, ly, lz, nx, ny, nz, scale):
        super().__init__()

        self.set_scale(scale)

        basevs = [
            Vertex(0, 0, 0, "v0"),
            Vertex(lx, 0, 0, "v1"),
            Vertex(lx, ly, 0, "v2"),
            Vertex(0, ly, 0, "v3"),
        ]

        for v in basevs:
            self.add_vertex(v.x, v.y, 0, v.name + "-bot")
            self.add_vertex(v.x, v.y, lz, v.name + "-top")

        vertex_names = [
            f"v{index}{post}" for post in ("-bot", "-top") for index in range(4)
        ]

        self.block = self.add_hexblock(vertex_names, (nx, ny, nz), name="")



"""OpenFOAM input files

.. rubric:: AST and parser

.. autosummary::
   :toctree:

    ast
    generators
    parser
    format

.. rubric:: Helper to create input files

.. autosummary::
   :toctree:

    blockmesh
    control_dict
    fields
    fv_schemes
    constant_files
    fv_options
    decompose_par
    util

"""
from abc import ABC, abstractmethod

from inflection import underscore

from .ast import Dict, DimensionSet, FoamInputFile, List, Value
from .parser import dump, parse

__all__ = [
    "parse",
    "dump",
    "create_field_from_code",
    "FoamInputFile",
    "DEFAULT_HEADER",
    "ControlDictHelper",
    "Dict",
    "List",
    "Value",
    "BlockMeshDict",
    "VolScalarField",
    "VolVectorField",
    "FvSchemesHelper",
    "Vertex",
    "read_field_file",
    "FileHelper",
    "ConstantFileHelper",
    "BlockMeshDictRectilinear",
    "FvOptionsHelper",
    "DimensionSet",
    "DecomposeParDictHelper",
    "format_code",
    "FoamFormatError",
]


def _as_py_name(name):
    return underscore(name).replace(".", "_")


def _update_dict_with_params(data, params_data):
    for key, value in data.items():
        name = underscore(key)

        try:
            param_value = params_data[name]
        except (KeyError, AttributeError):
            continue

        if isinstance(value, dict):
            _update_dict_with_params(value, param_value)
        else:
            data[key] = param_value


def _complete_params_dict(subparams, name, default, doc=None):
    name = _as_py_name(name)
    subsubparams = subparams._set_child(name, doc=doc)

    for key, value in default.items():
        if isinstance(value, dict):
            _complete_params_dict(subsubparams, key, value)
            continue

        subsubparams._set_attrib(_as_py_name(key), value)


class FileHelper(ABC):
    """Abstract class for "Helper" objects"""

    @abstractmethod
    def complete_params(self, params):
        """Complete the params object"""

    @abstractmethod
    def make_tree(self, params):
        """Make the AST corresponding to a file"""


DEFAULT_HEADER = r"""
/*--------------------------------*- C++ -*----------------------------------*\
| =========                 |                                                 |
| \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox           |
|  \\    /   O peration     | Version:  v2206                                 |
|   \\  /    A nd           | Website:  www.openfoam.com                      |
|    \\/     M anipulation  |                                                 |
\*---------------------------------------------------------------------------*/
"""

DEFAULT_HEADER = DEFAULT_HEADER[1:-1]

from .blockmesh import BlockMeshDict, BlockMeshDictRectilinear, Vertex
from .constant_files import ConstantFileHelper
from .control_dict import ControlDictHelper
from .decompose_par import DecomposeParDictHelper
from .fields import (
    VolScalarField,
    VolVectorField,
    create_field_from_code,
    read_field_file,
)
from .format import FoamFormatError, format_code
from .fv_options import FvOptionsHelper
from .fv_schemes import FvSchemesHelper



"""Subpackage containing potential templates (actually empty for now)

"""

import jinja2


class BaseTemplates:
    @property
    def env(self):
        return jinja2.Environment(
            loader=jinja2.PackageLoader("fluidsimfoam", "resources"),
            undefined=jinja2.StrictUndefined,
            keep_trailing_newline=True,
        )

    def get_base_template(self, name):
        """Get a template from ``fluidsimfoam.resources``."""
        return self.env.get_template(name)


_templates = BaseTemplates()

get_base_template = _templates.get_base_template



"""Helper to create fvSchemes files

"""

from inflection import underscore

from fluidsimfoam.foam_input_files import FileHelper, FoamInputFile
from fluidsimfoam.foam_input_files.util import as_dict


class FvSchemesHelper(FileHelper):
    keys = ["ddt", "grad", "div", "laplacian", "interpolation", "snGrad"]

    def __init__(
        self,
        ddt=None,
        grad=None,
        div=None,
        laplacian=None,
        interpolation=None,
        sn_grad=None,
    ):
        for key in self.keys:
            loc = locals()
            arg_name = underscore(key)
            if loc[arg_name] is None:
                loc[arg_name] = {}

            data = as_dict(loc[arg_name])
            setattr(self, arg_name, data)

        self.other_dict = {}

    def complete_params(self, params):
        fv_schemes = params._set_child("fv_schemes", doc="""TODO""")
        for key in self.keys:
            name = underscore(key)
            fv_schemes._set_child(name, attribs=getattr(self, name))

        for key, value in self.other_dict.items():
            name = underscore(key)
            fv_schemes._set_child(name, attribs=value)

    def make_tree(self, params):
        tree = FoamInputFile(
            info={
                "version": 2.0,
                "format": "ascii",
                "class": "dictionary",
                "object": "fvSchemes",
            }
        )

        def set_child(key, name=None):
            if name is None:
                name = key + "Schemes"
            params_key = params.fv_schemes[underscore(key)]
            attribs = {k: params_key[k] for k in params_key._key_attribs}
            attribs = {k: v for k, v in attribs.items() if v is not False}
            tree.set_child(name, attribs)

        for key in self.keys:
            set_child(key)

        for key in self.other_dict:
            set_child(key, name=key)

        return tree

    def add_dict(self, name, data):
        self.other_dict[name] = data



"""Helper to create fvOptions files

"""
import collections.abc

from fluidsimfoam.foam_input_files import (
    FileHelper,
    FoamInputFile,
    _as_py_name,
    _complete_params_dict,
    _update_dict_with_params,
)


def deep_update(d, u):
    # from https://stackoverflow.com/a/3233356
    for k, v in u.items():
        if isinstance(v, collections.abc.Mapping):
            d[k] = deep_update(d.get(k, {}), v)
        else:
            d[k] = v
    return d


def _make_default_params_dict(default: dict, name_parameters: set, result: dict):
    for name, value in default.items():
        if isinstance(value, dict):
            name_parameters1 = set()
            for full_name in name_parameters:
                if full_name.startswith(name + "/"):
                    name_parameters1.add(full_name[len(name) + 1 :])

            if name_parameters1:
                result[name] = {}
                _make_default_params_dict(value, name_parameters1, result[name])
        else:
            if name in name_parameters:
                result[name] = value

    # check if all name_parameters entries correspond to a parameter
    for full_name in name_parameters:
        keys = full_name.split("/")
        thing = default
        try:
            for key in keys:
                thing = thing[key]
        except KeyError:
            raise ValueError(
                f"'{full_name}' does not correspond to a parameter in {default}"
            )

    return result


def get_selection_mode(cell_zone, cell_set, points):
    is_not_none = [var is not None for var in (cell_zone, cell_set, points)]
    nb_var_not_none = sum(is_not_none)

    if nb_var_not_none == 0:
        return "all"
    elif nb_var_not_none > 1:
        raise ValueError

    if cell_zone is not None:
        selection_mode = "cellZone"
    elif cell_set is not None:
        selection_mode = "cellSet"
    elif points is not None:
        selection_mode = "points"
    else:
        raise RuntimeError

    return selection_mode


class FvOption:
    def __init__(
        self,
        type,
        name,
        active=True,
        cell_zone=None,
        cell_set=None,
        points=None,
        default=None,
        coeffs=None,
        parameters=None,
    ):
        self.type = type
        self.name = name

        if parameters is None:
            parameters = set()
        else:
            parameters = set(parameters)

        if default is not None and coeffs is not None:
            raise ValueError("default is not None and coeffs is not None")

        if default is None:
            default = {}

        default["type"] = type
        self.active = active

        self.coeffs = coeffs
        if coeffs is not None:
            default["coeffs"] = coeffs
            self.name_coeffs_key = type + "Coeffs"

        self.cell_zone = cell_zone
        self.cell_set = cell_set
        self.points = points

        self.default = default
        self.parameters = parameters

    def complete_params(self, params_fv_options):
        default_params = {
            "active": self.active,
            "selection": {
                "cell_zone": self.cell_zone,
                "cell_set": self.cell_set,
                "points": self.points,
            },
        }
        if self.coeffs is None:
            parameters = self.parameters
        else:
            parameters = set("coeffs/" + name for name in self.parameters)
        _make_default_params_dict(self.default, parameters, default_params)
        _complete_params_dict(params_fv_options, self.name, default_params)

    def get_dict_for_tree(self, params_fv_options):
        params_option = params_fv_options[_as_py_name(self.name)]

        dict_for_tree = {"type": self.type, "active": None}

        if self.coeffs is None:
            dict_select = dict_for_tree
        else:
            dict_select = dict_for_tree.setdefault("coeffs", {})

        select = params_option.selection
        dict_select["selectionMode"] = get_selection_mode(
            select.cell_zone, select.cell_set, select.points
        )

        for key in ("cellZone", "cellSet", "points"):
            key_as_py_name = _as_py_name(key)
            if select[key_as_py_name] is not None:
                dict_select[key] = select[key_as_py_name]

        deep_update(dict_for_tree, self.default)

        _update_dict_with_params(dict_for_tree, params_option)

        active = params_option.active
        if isinstance(active, bool):
            active = ("no", "yes")[int(active)]
        dict_for_tree["active"] = active
        if self.coeffs is not None:
            dict_for_tree[self.name_coeffs_key] = dict_for_tree.pop("coeffs")
        return dict_for_tree


class FvOptionsHelper(FileHelper):
    def __init__(self):
        self.options = {}

    def add_option(
        self,
        type,
        name=None,
        active=True,
        cell_zone=None,
        cell_set=None,
        points=None,
        default=None,
        coeffs=None,
        parameters=None,
    ):
        if name is None:
            name = type

        self.options[name] = FvOption(
            type,
            name,
            active,
            cell_zone,
            cell_set,
            points,
            default,
            coeffs,
            parameters,
        )

    def remove_option(self, name):
        del self.options[name]

    def complete_params(self, params):
        params_fv_options = params._set_child("fv_options", doc="""TODO""")
        for option in self.options.values():
            option.complete_params(params_fv_options)

    def make_tree(self, params):
        tree = FoamInputFile(
            info={
                "version": 2.0,
                "format": "ascii",
                "class": "dictionary",
                "location": '"system"',
                "object": "fvOptions",
            }
        )
        for name, option in self.options.items():
            tree.init_from_py_objects(
                {name: option.get_dict_for_tree(params.fv_options)}
            )
        return tree



"""Invoke tasks to be used from the ``tasks.py`` file in the solvers"""

import re
from pathlib import Path
from subprocess import Popen
from time import sleep, time

from invoke import task

from fluiddyn.util import time_as_str
from fluidsimfoam.foam_input_files import parse


@task
def clean(context):
    context.run("foamCleanTutorials", warn=True)


@task
def block_mesh(context):
    if (
        Path("system/blockMeshDict").exists()
        and not Path("constant/polyMesh").is_dir()
    ):
        context.run("blockMesh")


@task(block_mesh)
def polymesh(context):
    """Create the polymesh directory"""


@task
def set_fields(context, force=False):
    if Path("system/setFieldsDict").exists():
        path_setFields_called = Path(".data_fluidsim/setFields_called")
        path_setFields_called.parent.mkdir(exist_ok=True)
        if force or not path_setFields_called.exists():
            path_setFields_called.touch()
            context.run("setFields")


@task(polymesh, set_fields)
def run(context):
    """Main target to launch a simulation"""
    with open("system/controlDict") as file:
        ctr_dict = file.read()
    ctr_dict = parse(ctr_dict)

    application = ctr_dict.children["application"]
    end_time = ctr_dict["endTime"]
    start_time = ctr_dict["startTime"]

    path_run = Path.cwd()

    path_decompose_par_dict = path_run / "system/decomposeParDict"
    parallel = path_decompose_par_dict.exists()
    if parallel:
        context.run("decomposePar")
        nsubdoms = None
        with open(path_decompose_par_dict) as file:
            for line in file:
                if "numberOfSubdomains" in line:
                    line = line.strip().removesuffix(";")
                    nsubdoms = int(line.split()[-1])
                    break
        if nsubdoms is None:
            raise RuntimeError(f"Bad decomposeParDict {path_decompose_par_dict}")
        command = ["mpirun", "-np", str(nsubdoms), application, "-parallel"]
    else:
        command = application

    path_log = path_run / f"log_{time_as_str()}.txt"
    print(f"Starting simulation in \n{path_run}")
    with open(path_log, "w") as file_log:
        file_log.write(f"{start_time = }\n{end_time = }\n")
        file_log.flush()
        print(f"{end_time = }")
        pattern_time = re.compile(r"\nTime = ([\d]+\.[\d]+)")
        t_start = time()
        process = Popen(command, stdout=file_log)
        t_last = time() - 10.0
        while process.poll() is None:
            sleep(0.2)
            t_now = time()
            if t_now - t_last > 2:
                t_last = t_now
                log_size = path_log.stat().st_size
                with open(path_log, "r") as file_log_read:
                    file_log_read.seek(max(0, log_size - 1000))
                    log = file_log_read.read()
                if log:
                    groups = pattern_time.findall(log)
                    if groups:
                        time_simul = float(groups[-1])
                        print(
                            f"eq_time: {time_simul:12.3f} "
                            f"({100 * time_simul/end_time:6.2f} %), "
                            f"clock_time: {t_now - t_start: 12.3f} s"
                        )

    print(f"Simulation done. path_run:\n{path_run}")



"""Console script functions

"""

import argparse
import datetime
import re
import sys
from functools import partial
from importlib import resources
from pathlib import Path
from pprint import pprint
from shutil import copy
from string import Template

from inflection import camelize

from fluidsim_core.ipy_load import (
    start_ipython_load_sim as _start_ipython_load_sim,
)
from fluidsim_core.paths import path_dir_results
from fluidsimfoam import __version__ as fluidsimfoam_version
from fluidsimfoam.solvers import available_solvers


def print_versions():
    import fluiddyn
    import fluidsim_core
    import fluidsimfoam

    versions = {"Package": "Version", "-------": "-------"}

    packages = [fluidsimfoam, fluiddyn, fluidsim_core]
    for package in packages:
        versions[package.__name__] = package.__version__

    for pkg_name, version in versions.items():
        print(f"{pkg_name.ljust(15)} {version}")

    names = sorted(set([entry_point.name for entry_point in available_solvers()]))
    print("\nInstalled solvers: " + ", ".join(names))


start_ipython_load_sim = partial(
    _start_ipython_load_sim, load_import="from fluidsimfoam import load"
)


def initiate_solver():
    from fluidsimfoam.foam_input_files import FoamFormatError, format_code

    parser = argparse.ArgumentParser(
        prog="fluidsimfoam-initiate-solver",
        description="Initiate a new solver",
    )

    parser.add_argument("name")

    parser.add_argument("-c", "--from-case")
    parser.add_argument("-s", "--from-solver")
    parser.add_argument("-a", "--author", default="Fluidsimfoam developers")
    parser.add_argument("-d", "--destination")

    args = parser.parse_args()

    print(args)

    if args.from_case is None and args.from_solver is None:
        print("You should at least provide a case (-c) or a solver (-s)")
        sys.exit(1)

    if args.from_solver is not None:
        raise NotImplementedError("Sorry, option -s is not yet implemented...")

    if args.from_case is not None:
        path_case = Path(args.from_case)
        if not path_case.exists():
            print(f"{path_case} does not exist.")
            sys.exit(1)

    if args.destination is not None:
        path_destination = Path(args.destination).resolve()
    else:
        path_destination = path_dir_results

    name_project = f"fluidsimfoam-{args.name}"
    suffix_for_module = args.name.replace("-", "_")
    name_package = f"fluidsimfoam_{suffix_for_module}"
    path_result = path_destination / name_project

    if path_result.exists():
        print(f"{path_result} already exists.")
        sys.exit(1)

    print(f"Let's initiate the solver {path_result}")

    if args.from_case is not None:
        print(f"Using case {path_case}")

        path_files = {}
        name_files = {}

        for name_dir in ("system", "constant", "0", "0.orig"):
            path_files[name_dir] = sorted(
                path.relative_to(path_case)
                for path in path_case.glob(name_dir + "/*")
                # directories are not supported...
                if path.is_file()
            )

        paths_orig = path_files.pop("0.orig")
        if paths_orig:
            path_files["0"] = paths_orig

        for name_dir in path_files.keys():
            name_files[name_dir] = [p.name for p in path_files[name_dir]]

        pprint(name_files)

    print(f"Making {path_result}")
    path_result.mkdir()

    substitutions = {
        "year": str(datetime.date.today().year),
        "author": args.author,
        "name_short": args.name,
        "name_project": name_project,
        "name_package": name_package,
        "fluidsimfoam_version": fluidsimfoam_version,
        "suffix_for_class": camelize(suffix_for_module),
        "name_variables": str(name_files["0"]),
        "name_system_files": str(name_files["system"]),
        "name_constant_files": str(name_files["constant"]),
    }

    templates = {}

    for name in ["LICENSE", "README.md", "pyproject.toml"]:
        relative_path = name
        templates[relative_path] = name + ".template"

    path_package = path_result / f"src/{name_package}"
    for name in ["__init__.py", "output.py", "tasks.py"]:
        relative_path = path_package / name
        templates[relative_path] = name + ".template"

    path_templates = path_package / "templates"
    path_templates.mkdir(parents=True)
    name = "tasks.py"
    templates[path_templates / name] = name + ".template"

    path_tests = path_result / "tests"
    path_tests.mkdir()
    templates[
        path_tests / f"test_{suffix_for_module}.py"
    ] = "test_generated_solver.py.template"

    for relative_path, template in templates.items():
        template_res = resources.files("fluidsimfoam.resources").joinpath(
            template
        )
        with resources.as_file(template_res) as path:
            code = path.read_text()
        code = Template(code).substitute(substitutions)
        (path_result / relative_path).write_text(code)

    path_saved_case = path_tests / "saved_cases/case0"
    path_saved_case.mkdir(parents=True)

    for name_dir, path_files_dir in path_files.items():
        as_field = name_dir == "0"
        path_dir_saved_case = path_saved_case / name_dir
        path_dir_saved_case.mkdir()

        for relative_path in path_files_dir:
            code = (path_case / relative_path).read_text()
            try:
                code = format_code(code, as_field=as_field)
            except FoamFormatError:
                print(f"Not able to format file {relative_path}")

            # Trim trailing whitespaces
            code = re.sub(r"\s+\n", "\n", code).strip() + "\n"

            if "{{" in code:
                # Escape jinja syntax
                code_template = "{% raw %}" + code.strip() + "{% endraw %}\n"
            else:
                code_template = code

            if name_dir == "0":
                parts = list(relative_path.parts)
                if parts[0] == "0.orig":
                    parts[0] = "0"
                    relative_path = Path(*parts)

            (path_saved_case / relative_path).write_text(code)
            (path_templates / (relative_path.name + ".jinja")).write_text(
                code_template
            )

    if args.from_case is not None:
        for name in ("Allclean", "Allrun"):
            path = path_case / name
            if path.exists():
                copy(path, path_saved_case)

    print(
        f"""

New solver {name_project} created! You can now install and test it to check if
it reproduces your initial case.

```
cd {path_result}
pip install -e .
pytest tests -vv
```

You can then improve your fluidsimfoam solver by modifying its files
(especially the file src/{name_package}/output.py). Examples can be found in
https://foss.heptapod.net/fluiddyn/fluidsimfoam/-/tree/branch/default/doc/examples

    """
    )



"""Parser for OpenFOAM input files"""

from dataclasses import dataclass
from pathlib import Path

from lark import Lark, Token, Transformer
from lark.exceptions import LarkError

from .ast import (
    Assignment,
    Code,
    CodeStream,
    Dict,
    DimensionSet,
    Directive,
    FoamInputFile,
    List,
    Name,
    Value,
    VariableAssignment,
)

here = Path(__file__).absolute().parent

grammar = (here / "grammar.lark").read_text()
grammar_advanced = (here / "grammar_advanced.lark").read_text()

lark_parser = Lark(grammar, start="value", lexer="basic")
lark_parser_advanced = Lark(grammar_advanced, start="value", lexer="basic")

parsers = {"simple": lark_parser, "advanced": lark_parser_advanced}


@dataclass
class ListInfo:
    name: str
    info: str = None
    dtype: str = None
    size: int = None


def parse(text, grammar=None):
    text = "\n".join(line.rstrip() for line in text.split("\n"))
    if not text.endswith("\n"):
        text += "\n"

    if grammar is None:
        try:
            tree = lark_parser.parse(text)
        except LarkError:
            tree = lark_parser_advanced.parse(text)
    else:
        tree = parsers[grammar].parse(text)

    return FoamTransformer().transform(tree)


def dump(tree):
    return tree.dump()


def filter_no_newlines(items):
    return [item for item in items if item is not None]


def _convert_to_number(number):
    number = str(number)
    try:
        return int(number)
    except ValueError:
        return float(number)


class FoamTransformer(Transformer):
    def SIGNED_NUMBER(self, token):
        return _convert_to_number(token)

    def CNAME(self, token):
        return token.value

    def NEWLINE(self, nodes):
        return None

    def ESCAPED_STRING(self, token):
        return token.value

    def DOUBLE_NAME(self, token):
        return token.value

    def TRIPLE_NAME(self, token):
        return token.value

    def EQKEY(self, token):
        return token.value

    def MACRO(self, token):
        return token.value

    def dimension_set(self, items):
        return DimensionSet(
            [
                item
                for item in items
                if not (isinstance(item, Token) and item.type == "NEWLINE")
            ]
        )

    def directive(self, nodes):
        if len(nodes) != 1:
            raise RuntimeError
        return str(nodes[0])

    def list(self, items):
        return List(
            [
                item
                for item in items
                if not (isinstance(item, Token) and item.type == "NEWLINE")
                and item is not None
            ]
        )

    def file(self, nodes):
        first_assignment = nodes[0]
        if first_assignment.name == "FoamFile":
            info_dict = first_assignment.value
            nodes = nodes[1:]
        else:
            info_dict = None

        for node in nodes:
            if isinstance(node.value, List):
                node.value.add_name(node.name)

        return FoamInputFile(info_dict, {node.name: node.value for node in nodes})

    def var_assignment(self, nodes):
        nodes = [node for node in nodes if node is not None]
        dimension_set = None
        name_in_value = None
        name = nodes.pop(0)
        if len(nodes) == 1:
            value = nodes[0]
        else:
            try:
                index_dimension = [
                    isinstance(elem, DimensionSet) for elem in nodes
                ].index(True)
            except ValueError:
                dimension_set = None
            else:
                dimension_set = nodes.pop(index_dimension)

            name_in_value = None
            if all([isinstance(elem, str) for elem in nodes]):
                last_value = " ".join(nodes)
            else:
                last_value = nodes.pop(-1)
                if not all(isinstance(node, str) for node in nodes):
                    nodes = [str(node) for node in nodes]
                if nodes:
                    name_in_value = " ".join(nodes)
                if isinstance(last_value, List):
                    last_value._name = name_in_value

            if (name_in_value is None and dimension_set is None) or isinstance(
                last_value, List
            ):
                value = last_value
            else:
                value = Value(
                    last_value, name=name_in_value, dimension=dimension_set
                )
        return VariableAssignment(name, value)

    def dict_assignment(self, nodes):
        nodes = filter_no_newlines(nodes)
        directive = None

        if len(nodes) == 1:
            name = nodes.pop(0)
            return Assignment(name, Dict(data={}, name=name))

        nodes_str = []
        for node in nodes:
            if isinstance(node, str) and node.startswith("#"):
                # like #codeStream
                directive = nodes.pop(1)
                break
        if isinstance(nodes[0], list):
            name = "(" + " ".join(nodes.pop(0)) + ")"
        else:
            name = nodes.pop(0)
        if nodes_str:
            name += " " + " ".join(nodes_str)

        nodes_assign = [
            node
            for node in nodes
            if hasattr(node, "name") and hasattr(node, "value")
        ]

        for node in nodes_assign:
            if isinstance(node.value, List):
                node.value.add_name(node.name)

        if directive is not None and directive == "#codeStream":
            cls = CodeStream
        else:
            cls = Dict

        return Assignment(
            name,
            cls(
                data={node.name: node.value for node in nodes_assign},
                name=name,
                directive=directive,
            ),
        )

    def isolated_list(self, nodes):
        nodes = filter_no_newlines(nodes)
        if len(nodes) != 1:
            raise NotImplementedError(nodes)
        name = None
        return Assignment(name, nodes[0])

    def list_info(self, nodes):
        nodes = filter_no_newlines(nodes)

        dtype = None
        for inode, node in enumerate(nodes):
            if isinstance(node, Token) and node.type == "LIST_TYPE":
                dtype = str(node)[5:-1]
                break
        if dtype is not None:
            nodes.pop(inode)

        size = None
        if len(nodes) > 1 and isinstance(nodes[-1], int):
            size = nodes.pop(-1)

        name = nodes.pop(0)

        info = None
        if nodes:
            info = " ".join(nodes)

        return ListInfo(name=name, info=info, dtype=dtype, size=size)

    def list_assignment(self, nodes):
        nodes = filter_no_newlines(nodes)

        list_info, the_list = nodes
        assert isinstance(list_info, ListInfo)

        name = list_info.name
        name_internal = name

        if list_info.info is not None:
            name_internal += " " + list_info.info

        if list_info.dtype is not None:
            name_internal += f"\nList<{list_info.dtype}>"

        if list_info.size is not None:
            name_internal += f"\n{list_info.size}"

        if isinstance(name, int):
            name = str(name)
        if isinstance(name_internal, int):
            name_internal = str(name_internal)

        the_list._name = name_internal
        return Assignment(name, the_list)

    def dimension_assignment(self, nodes):
        nodes = [node for node in nodes if node is not None]
        name = nodes.pop(0)

        if len(nodes) == 3:
            return Assignment(name, Value(nodes[-1], nodes[0], nodes[-2]))
        elif len(nodes) == 2:
            return Assignment(name, Value(nodes[-1], dimension=nodes[-2]))
        else:
            raise RuntimeError(nodes)

    def macro_assignment(self, nodes):
        nodes = [node for node in nodes if node is not None]
        if len(nodes) != 1:
            raise NotImplementedError(nodes)
        name = nodes.pop(0)
        return Assignment(name, "")

    def equal_assign(self, nodes):
        nodes = [node for node in nodes if node is not None]
        if len(nodes) != 2:
            raise RuntimeError(nodes)
        value = nodes[1]
        if hasattr(value, "dump"):
            value = value.dump()
        return f"{nodes[0]}={value}"

    def directive_assignment(self, nodes):
        nodes = [node for node in nodes if node is not None]
        if len(nodes) == 2:
            directive, content = nodes
        else:
            directive = nodes.pop(0)
            function_name = nodes.pop(0)
            arguments = ", ".join(nodes)
            content = function_name + f"({arguments})"
        key = directive + " " + content
        return Assignment(key, Directive(directive, content))

    def code_assignment(self, nodes):
        nodes = filter_no_newlines(nodes)
        directive = None
        if len(nodes) == 2:
            name, code = nodes
        elif len(nodes) == 3:
            name, directive, code = nodes
            if not directive.startswith("#"):
                raise NotImplementedError(nodes)
        else:
            raise NotImplementedError(nodes)
        code = str(code)
        if code.startswith("#{\n"):
            code = code.split("\n", 1)[-1]
            code = code.rsplit("\n", 1)[0]
        else:
            code = code[2:-3].strip()
        return Assignment(name, Code(name, code, directive=directive))

    def special_directives(self, token):
        token = filter_no_newlines(token)
        return Name(token[0].value + "\n")



"""Utilities"""
from typing import Union


def as_dict(data: Union[dict, str], filter_comments=True) -> dict:
    """return a dict, potentially by converting a str"""
    if isinstance(data, dict):
        return data
    elif isinstance(data, str):
        data_as_str = data
        data = {}
        for line in data_as_str.strip().split("\n"):
            line = line.strip().removesuffix(";")
            if any(line.startswith(comment_char) for comment_char in ("//", "#")):
                if filter_comments:
                    continue
                else:
                    key, value = line, None
            else:
                key, value = line.split(maxsplit=1)
            data[key] = value
        return data
    else:
        raise TypeError



"""Base class for the ``sim`` object"""

from pathlib import Path
from time import sleep, time

from inflection import underscore

from fluiddyn.util import mpi  # noqa: F401
from fluidsim_core.solver import SimulCore
from fluidsimfoam.log import logger
from fluidsimfoam.params import Parameters


class SimulFoam(SimulCore):
    """Base OpenFOAM Fluidsim solver."""

    Parameters = Parameters

    @classmethod
    def _complete_params_with_default(cls, params):
        pass

    @classmethod
    def create_default_params(cls):
        # TODO: needed because abstract method in fluidsim-core=<0.7.2
        params = super().create_default_params()
        return params

    def __init__(self, params):
        super().__init__(params)

        self._objects_to_print = "{:28s}{}\n".format("sim: ", type(self))
        dict_classes = self.info_solver.import_classes()

        # initialize objects
        for cls_name, Class in dict_classes.items():
            # only initialize if Class is not the Simul class
            if isinstance(self, Class):
                continue

            obj_name = underscore(cls_name)
            setattr(self, obj_name, Class(self))
            self._objects_to_print += "{:28s}{}\n".format(
                f"sim.{obj_name}: ", Class
            )

        if "Output" in dict_classes:
            if not params.NEW_DIR_RESULTS:
                self.path_run = self.output.path_run = Path(params.path_run)
            else:
                # path_run would be initialized by the Output instance if available
                # See self.output._init_name_run()
                self.path_run = Path(self.output.path_run)

            self.output.post_init()
        else:
            self.path_run = None
            if mpi.rank == 0:
                logger.warning("No output class initialized!")

        self.input_files = self.output.input_files

    def stop_time_loop(self, stop_at="writeNow", verbose=True):
        if verbose:
            print("Telling OpenFOAM to stop... (overwrite controlDict file)")
            t_stop = time()
        ctr_dict = self.input_files.control_dict.read()
        ctr_dict["stopAt"] = stop_at
        ctr_dict.overwrite()

        process = self.make.process

        if process is None:
            return

        while process.poll() is None:
            # touch needed (strange OpenFOAM bug?)
            (self.path_run / "system/controlDict").touch()
            sleep(0.05)
            if verbose:
                saved_directories = sorted(
                    path.name
                    for path in self.path_run.glob("*")
                    if path.name[0].isdigit()
                )
                print(f"\rsaved times {saved_directories}", end="")

        if verbose:
            print(
                f"\nSimulation stopped {time() - t_stop:.2f} s "
                f"after `stopAt = {stop_at}`"
            )



"""Create a dataframe from result directories"""

from fluidsim_core.output.dataframe_from_paths import DataframeMaker
from fluidsimfoam import load_simul
from fluidsimfoam.output.log import read_time_last


class DataframeMakerFoam(DataframeMaker):
    def get_time_start_from_path(self, path):
        """Get first time"""
        logfiles = sorted(path.glob("log*.txt"))
        path_log = logfiles[0]
        with open(path_log) as file:
            for line in file:
                if line.startswith("start_time = "):
                    return float(line.split()[-1])

    def get_time_last_from_path(self, path):
        """Get last saved time"""
        logfiles = sorted(path.glob("log*.txt"))
        path_log = logfiles[-1]
        return read_time_last(path_log)

    def load_sim(self, path):
        """Load a simulation object"""
        return load_simul(path, hide_stdout=True)


_dataframe_maker = DataframeMakerFoam()

get_mean_values_from_path = _dataframe_maker.get_mean_values_from_path
get_dataframe_from_paths = _dataframe_maker.get_dataframe_from_paths



"""Base class for the ``sim.output`` object

.. autosummary::
   :toctree:

   base
   fields
   log
   dataframe_from_paths

"""

from fluidsim_core.paths import path_dir_results

from .base import Output
from .dataframe_from_paths import (
    get_dataframe_from_paths,
    get_mean_values_from_path,
)

__all__ = [
    "path_dir_results",
    "Output",
    "get_dataframe_from_paths",
    "get_mean_values_from_path",
]



from collections.abc import Iterable
from itertools import product


class Grading:
    """base class for Simple- and Edge- Grading"""


class SimpleGradingElement:
    """x, y or z Element of simpleGrading. adopted to multi-grading"""

    def __init__(self, d):
        """initialization
        d is single number for expansion ratio
          or iterative object consits (direction ratio, cell ratio, expansion ratio)
        """
        self.d = d

    def format(self):
        if isinstance(self.d, Iterable):
            return (
                "( "
                + " ".join(f"( {e[0]:g} {e[1]:g} {e[2]:g} )" for e in self.d)
                + " )"
            )
        else:
            return str(self.d)


class SimpleGrading(Grading):
    """configutation for 'simpleGrading'"""

    def __init__(self, x, y, z):
        if not isinstance(x, SimpleGradingElement):
            self.x = SimpleGradingElement(x)
        else:
            self.x = x
        if not isinstance(y, SimpleGradingElement):
            self.y = SimpleGradingElement(y)
        else:
            self.y = y
        if not isinstance(z, SimpleGradingElement):
            self.z = SimpleGradingElement(z)
        else:
            self.z = z

    def format(self):
        return (
            f"simpleGrading "
            f"({self.x.format()} {self.y.format()} {self.z.format()})"
        )


class EdgeGrading(Grading):
    """configutation for 'edgeGrading'"""

    def __init__(self, x1, x2, x3, x4, y1, y2, y3, y4, z1, z2, z3, z4):
        if not isinstance(x1, SimpleGradingElement):
            self.x1 = SimpleGradingElement(x1)
        else:
            self.x1 = x1
        if not isinstance(x2, SimpleGradingElement):
            self.x2 = SimpleGradingElement(x2)
        else:
            self.x2 = x2
        if not isinstance(x3, SimpleGradingElement):
            self.x3 = SimpleGradingElement(x3)
        else:
            self.x3 = x3
        if not isinstance(x4, SimpleGradingElement):
            self.x4 = SimpleGradingElement(x4)
        else:
            self.x4 = x4
        if not isinstance(y1, SimpleGradingElement):
            self.y1 = SimpleGradingElement(y1)
        else:
            self.y1 = y1
        if not isinstance(y2, SimpleGradingElement):
            self.y2 = SimpleGradingElement(y2)
        else:
            self.y2 = y2
        if not isinstance(y3, SimpleGradingElement):
            self.y3 = SimpleGradingElement(y3)
        else:
            self.y3 = y3
        if not isinstance(y4, SimpleGradingElement):
            self.y4 = SimpleGradingElement(y4)
        else:
            self.y4 = y4
        if not isinstance(x1, SimpleGradingElement):
            self.z1 = SimpleGradingElement(z1)
        else:
            self.z1 = z1
        if not isinstance(z2, SimpleGradingElement):
            self.z2 = SimpleGradingElement(z2)
        else:
            self.z2 = z2
        if not isinstance(z3, SimpleGradingElement):
            self.z3 = SimpleGradingElement(z3)
        else:
            self.z3 = z3
        if not isinstance(z4, SimpleGradingElement):
            self.z4 = SimpleGradingElement(z4)
        else:
            self.z4 = z4

    def format(self):
        tmp = []
        for letter, index in product("xyz", "1234"):
            var = getattr(self, letter + index)
            tmp.append(var.format())
        return "edgeGrading (" + " ".join(tmp) + ")"



"""Class for the ``sim.output.fields`` object"""

import shutil
from numbers import Number
from subprocess import PIPE, run

import matplotlib.pyplot as plt

from fluidsimfoam.foam_input_files import read_field_file


def is_time_name(name):
    return all(c.isdigit() or c == "." for c in name)


class Fields:
    def __init__(self, output):
        self.output = output
        self.sim = output.sim

    def get_saved_times(self):
        if self.sim.params.parallel.nsubdoms > 1:
            str_glob = "processor0/*"
        else:
            str_glob = "*"

        return sorted(
            float(path.name)
            for path in self.output.path_run.glob(str_glob)
            if is_time_name(path.name)
        )

    def get_path_dir_time(self, time_approx="last", dirname=None):
        if time_approx != "last":
            raise NotImplementedError

        if dirname is None:
            str_glob = "*"
        else:
            str_glob = dirname + "/*"

        path_times = sorted(
            (
                path
                for path in self.output.path_run.glob(str_glob)
                if path.name[0].isdigit()
            ),
            key=lambda p: float(p.name),
        )
        path_dir = path_times[-1]
        last_time = float(path_dir.name)
        return path_dir, last_time

    def read_field(self, name, time_approx="last"):
        if time_approx != "last":
            raise NotImplementedError

        path_dir, last_time = self.get_path_dir_time(time_approx)

        if self.sim.params.parallel.nsubdoms > 1:
            _, last_time_proc0 = self.get_path_dir_time(
                time_approx, dirname="processor0"
            )

            if last_time_proc0 != last_time:
                self.reconstruct_par(fields=[name], time=last_time_proc0)

        path_dir, last_time = self.get_path_dir_time(time_approx)
        if self.sim.params.parallel.nsubdoms > 1:
            assert last_time == last_time_proc0

        field = read_field_file(path_dir / name)
        field.time = float(path_dir.name)
        return field

    def reconstruct_par(self, fields=None, latest_time=None, time=None):
        path_command = shutil.which("reconstructPar")

        if path_command is None:
            raise RuntimeError("OpenFOAM not available")

        command = ["reconstructPar"]

        if fields is not None:
            command.extend(["-fields", f"({' '.join(fields)})"])

        if latest_time is not None and time is not None:
            raise ValueError

        if latest_time is not None:
            command.append("-latestTime")

        if time is not None:
            if isinstance(time, Number):
                time = str(time)

            command.extend(["-time", time])

        run(command, cwd=self.sim.path_run, stdout=PIPE)

    def plot_field(self, name, time_approx="last"):
        field = self.read_field(name, time_approx)

        x, y, z = self.output.sim.oper.get_cells_coords()

        fig, ax = plt.subplots()

        ax.plot(y, field.get_array())



"""Base class for the ``sim.make`` object

"""


from subprocess import PIPE, Popen, run


class MakeInvoke:
    def __init__(self, sim=None):
        self.sim = sim
        self.process = None

    def exec(self, task_name, stdout=None):
        self.process = run(
            ["inv", task_name], cwd=self.sim.path_run, stdout=stdout
        )
        self.process.check_returncode()

    def list(self):
        self.exec("--list")

    def exec_async(self, task_name):
        self.process = Popen(
            ["inv", task_name], cwd=self.sim.path_run, stdout=PIPE
        )
        return self.process



"""Helper to create controlDict files"""

from inflection import underscore

from fluidsimfoam.foam_input_files import (
    DEFAULT_HEADER,
    FileHelper,
    FoamInputFile,
    _complete_params_dict,
)
from fluidsimfoam.foam_input_files.util import as_dict

# taken from https://doc.cfd.direct/openfoam/user-guide-v6/controldict
DEFAULT_CONTROL_DICT = dict(
    application="icoFoam",
    startFrom="startTime",
    startTime=0,
    stopAt="endTime",
    endTime=0.5,
    deltaT=0.005,
    writeControl="timeStep",
    writeInterval=20,
    purgeWrite=0,
    writeFormat="ascii",
    writePrecision=6,
    writeCompression="off",
    timeFormat="general",
    timePrecision=6,
    runTimeModifiable="true",
)


class Function:
    def __init__(self, type, libs, entries=None):
        self.type = type
        self.libs = libs
        self.entries = entries

    def make_dict(self):
        data = {"type": self.type, "libs": self.libs}
        if self.entries is not None:
            data.update(as_dict(self.entries, filter_comments=False))
        return data


class ControlDictHelper(FileHelper):
    def __init__(self, default=None):
        self.default_control_dict_params = DEFAULT_CONTROL_DICT.copy()
        if default is not None:
            self.default_control_dict_params.update(as_dict(default))

        self.functions = {}
        self.functions_included = {}

    def complete_params(self, params):
        _complete_params_dict(
            params, "control_dict", self.default_control_dict_params
        )

    def make_tree(self, params):
        tree = FoamInputFile(
            info={
                "version": "2.0",
                "format": "ascii",
                "class": "dictionary",
                "location": '"system"',
                "object": "controlDict",
            },
            children={
                key: params.control_dict[underscore(key)]
                for key in self.default_control_dict_params.keys()
            },
            header=DEFAULT_HEADER,
        )

        data = {}
        for name, kind in self.functions_included.items():
            if kind is None:
                kind = "#includeFunc"
            data[f"{kind} {name}"] = None
        for name, function in self.functions.items():
            data[name] = function.make_dict()
        if data:
            tree.init_from_py_objects({"functions": data})

        return tree

    def new(self, default=None):
        return type(self)(default)

    def add_function(self, key, type, libs, entries=None):
        self.functions[key] = Function(type, libs, entries)

    def include_function(self, name, kind=None):
        self.functions_included[name] = kind

    def include_functions(self, names, kind=None):
        for name in names:
            self.include_function(name, kind)



class ArcEdge:
    def __init__(self, vnames, name, inter_vertex):
        """Initialize ArcEdge instance
        vnames is the vertex names in order descrived in
          http://www.openfoam.org/docs/user/mesh-description.php
        # two vertices is needed for Arc
        cells is number of cells devied into in each direction
        name is the uniq name of the block
        grading is grading method.
        """
        self.vnames = vnames
        self.name = name
        self.inter_vertex = inter_vertex

    def format(self, vertices):
        """Format instance to dump
        vertices is dict of name to Vertex
        """
        index = " ".join(str(vertices[vn].index) for vn in self.vnames)
        comment = " ".join(self.vnames)
        v = self.inter_vertex
        return (
            f"arc {index} ({v.x:18.15g} {v.y:18.15g} {v.z:18.15g}) "
            f"// {self.name} ({comment})"
        )


class SplineEdge:
    def __init__(self, vnames, name, points):
        """Initialize SplineEdge instance
        vnames is the vertex names in order descrived in
          http://www.openfoam.org/docs/user/mesh-description.php
        # two vertices is needed for Spline
        """
        self.vnames = vnames
        self.name = name
        self.points = points

    def format(self, vertices):
        """Format instance to dump
        vertices is dict of name to Vertex
        """
        index = " ".join(str(vertices[vn].index) for vn in self.vnames)
        comment = " ".join(self.vnames)
        tmp = []
        tmp.append(
            f"spline {index}                      // {self.name} ({comment})"
        )
        tmp.append("    (")
        for p in self.points:
            tmp.append("         " + p.format())
        tmp.append(")")
        return "\n".join(tmp)



"""Base output class"""

import importlib.resources
import inspect
import os
import shutil
import textwrap
from contextlib import nullcontext
from pathlib import Path

from inflection import underscore

from fluiddyn.io import stdout_redirected
from fluiddyn.util import mpi
from fluidsim_core.output import OutputCore
from fluidsim_core.params import iter_complete_params
from fluidsimfoam.foam_input_files import (
    ConstantFileHelper,
    ControlDictHelper,
    DecomposeParDictHelper,
)
from fluidsimfoam.foam_input_files.generators import (
    InputFiles,
    new_file_generator_class,
)
from fluidsimfoam.log import logger
from fluidsimfoam.solvers import get_solver_package


def copy_resources(resources, path_run):
    if resources is None:
        return
    if isinstance(resources, str):
        resources = [resources]

    for resource in resources:
        resource = str(resource)
        if "->" in resource:
            resource, relative_path = resource.split("->")
            destination = path_run / relative_path.strip()
            resource = resource.strip()
        else:
            destination = path_run

        if resource.startswith("package-data("):
            resource = resource.removeprefix("package-data(")
            package_name, relative_path = resource.split(")")
            relative_path = relative_path.removeprefix("/")
            multiplexed_path = importlib.resources.files(package_name)
            resource = multiplexed_path.joinpath(relative_path)
            context_manager = importlib.resources.as_file
        else:
            resource = Path(os.path.expandvars(resource)).expanduser()
            context_manager = nullcontext

        with context_manager(resource) as resource:
            if resource.is_file():
                shutil.copy(resource, destination)
            elif resource.is_dir():
                shutil.copytree(resource, destination / resource.name)
            else:
                raise NotImplementedError(
                    f"{resource = }, {resource.exists() = }"
                )


class Output(OutputCore):
    name_variables = ["p", "U"]
    name_constant_files = ["transportProperties", "turbulenceProperties"]
    name_system_files = [
        "controlDict",
        "fvSchemes",
        "fvSolution",
        "decomposeParDict",
    ]

    _helper_turbulence_properties = ConstantFileHelper(
        "turbulenceProperties", {"simulationType": "laminar"}
    )
    _helper_decompose_par_dict = DecomposeParDictHelper()
    _helper_control_dict = ControlDictHelper()

    @classmethod
    def _complete_info_solver(cls, info_solver):
        """Complete the info_solver instance with child class details (module
        and class names).

        """
        classes = info_solver.classes.Output._set_child("classes")
        cls._set_info_solver_classes(classes)
        # iteratively call _complete_info_solver of the above classes
        info_solver.classes.Output.complete_with_classes()

    @classmethod
    def _set_info_solver_classes(cls, classes):
        """Set the classes for info_solver.classes.Output"""
        classes._set_child(
            "Log",
            dict(
                module_name="fluidsimfoam.output.log",
                class_name="Log",
            ),
        )

        classes._set_child(
            "Fields",
            dict(
                module_name="fluidsimfoam.output.fields",
                class_name="Fields",
            ),
        )

    @classmethod
    def _setup_file_generators_classes(cls):
        classes = cls._file_generators_classes = {}
        for variable_name in cls.name_variables:
            classes[variable_name.replace(".", "_")] = new_file_generator_class(
                variable_name
            )

        for file_name in cls.name_constant_files:
            classes[file_name] = new_file_generator_class(file_name, "constant")

        for file_name in cls.name_system_files:
            classes[file_name] = new_file_generator_class(file_name, "system")

    @classmethod
    def _complete_params_with_default(cls, params, info_solver):
        """This static method is used to complete the *params* container."""

        cls._setup_file_generators_classes()
        iter_complete_params(
            params, info_solver, cls._file_generators_classes.values()
        )

        # Bare minimum
        params._set_attribs(
            dict(NEW_DIR_RESULTS=True, short_name_type_run="run", resources=None)
        )
        params._set_child(
            "output", attribs={"HAS_TO_SAVE": True, "sub_directory": ""}
        )
        params.output._set_doc(
            textwrap.dedent(
                """
    - ``HAS_TO_SAVE``: bool (default: True) If False, nothing new is saved in
      the directory of the simulation.
    - ``sub_directory``: str (default: "") A name of a sub-directory (relative
      to $FLUIDDYN_PATH_SCRATCH) wherein the directory of the simulation
      (``path_run``) is saved.

"""
            )
        )

        dict_classes = info_solver.classes.Output.import_classes()
        iter_complete_params(params, info_solver, dict_classes.values())

    def __init__(self, sim=None, params=None):
        self.sim = sim
        try:
            self.name_solver = sim.info.solver.short_name
        except AttributeError:
            pass
        else:
            self.package = get_solver_package(self.name_solver)

        self.path_solver_package = self.get_path_solver_package()

        if sim:
            # self.oper = sim.oper
            self.params = sim.params.output
            super().__init__(sim)
        elif params:
            # At least initialize params
            self.params = params.output
        else:
            self.params = None
            logger.warning(
                "Initializing Output class without sim or params might lead to errors."
            )

        if not sim:
            return

        self.path_run = Path(self.path_run)

        self.input_files = InputFiles(self)
        # initialize objects
        dict_classes = sim.info.solver.classes.Output.import_classes()
        for cls_name, Class in dict_classes.items():
            if isinstance(self, Class):
                continue
            obj_name = underscore(cls_name)
            self.sim._objects_to_print += "{:28s}{}\n".format(
                f"sim.output.{obj_name}: ", Class
            )
            setattr(self, obj_name, Class(self))

        if not hasattr(self, "_file_generators_classes"):
            self._setup_file_generators_classes()
        for_str_input_files = {}
        for cls_name, Class in self._file_generators_classes.items():
            obj_name = underscore(cls_name)

            if Class.dir_name not in for_str_input_files:
                for_str_input_files[Class.dir_name] = [cls_name]
            else:
                for_str_input_files[Class.dir_name].append(cls_name)
            setattr(self.input_files, obj_name, Class(self))

        if for_str_input_files:
            self.sim._objects_to_print += "input_files:\n"
            for dir_name, input_files in for_str_input_files.items():
                self.sim._objects_to_print += (
                    f"  - in {dir_name + ':':12s}{' '.join(input_files)}\n"
                )

        if not (
            mpi.rank == 0
            and self._has_to_save
            and self.sim.params.NEW_DIR_RESULTS
        ):
            return

        (self.path_run / "system").mkdir()
        (self.path_run / "0").mkdir()
        (self.path_run / "constant").mkdir()

        for file_generator in vars(self.input_files).values():
            if hasattr(
                file_generator, "generate_file"
            ) and file_generator.rel_path.startswith("system"):
                file_generator.generate_file()

    @classmethod
    def get_path_solver_package(cls):
        """Get the path towards the solver package."""
        return Path(inspect.getmodule(cls).__file__).parent

    def post_init(self):
        """Logs info on instantiated classes"""

        if mpi.rank == 0:
            print(f"path_run: {self.path_run}")

        # This also calls _save_info_solver_params_xml
        with stdout_redirected():
            super().post_init()

        logger.info(self.sim._objects_to_print)

        # Write input files of the simulation
        if not (
            mpi.rank == 0
            and self._has_to_save
            and self.sim.params.NEW_DIR_RESULTS
        ):
            return

        copy_resources(self.sim.params.resources, self.path_run)
        self._post_init_create_additional_source_files()

        # OpenFOAM cleanup removes .xml files!
        path_info_fluidsim = self.path_run / ".data_fluidsim"
        path_info_fluidsim.mkdir(exist_ok=True)
        for file_name in ("info_solver.xml", "params_simul.xml"):
            path_new = path_info_fluidsim / file_name
            if path_new.exists():
                continue
            shutil.copy(self.path_run / file_name, path_new)

    def _post_init_create_additional_source_files(self):
        """Create the files from their template"""
        path_tasks_py = self.input_files.templates_dir / "tasks.py"
        if not path_tasks_py.exists():
            raise RuntimeError(
                "tasks.py missing in solver templates_dir "
                f"{self.input_files.templates_dir}"
            )
        shutil.copy(path_tasks_py, self.sim.path_run)

        for file_generator in vars(self.input_files).values():
            if hasattr(
                file_generator, "generate_file"
            ) and not file_generator.rel_path.startswith("system"):
                # system files are already generated
                file_generator.generate_file()

    @classmethod
    def _complete_params_block_mesh_dict(cls, params):
        default = {
            "nx": 40,
            "ny": 40,
            "nz": 40,
            "lx": 1.0,
            "ly": 1.0,
            "lz": 1.0,
            "scale": 1,
        }

        params._set_child(
            "block_mesh_dict",
            attribs=default,
            doc="""TODO""",
        )

    def _compute_mean_values(self, tmin, tmax):
        return {}



"""Helper to create constant files"""

from copy import deepcopy

from fluidsimfoam.foam_input_files import (
    DimensionSet,
    FileHelper,
    FoamInputFile,
    _as_py_name,
    _complete_params_dict,
    _update_dict_with_params,
)


class ConstantFileHelper(FileHelper):
    def __init__(
        self,
        file_name: str,
        default: dict,
        default_dimension: str = False,
        dimensions: dict = None,
        comments: dict = None,
        doc: str = None,
        cls: str = "dictionary",
        dimension=None,
    ):
        self.file_name = file_name
        self.default = default
        self.default_dimension = default_dimension
        self.dimensions = dimensions
        self.comments = comments
        self.doc = doc
        self.cls = cls
        self.dimension = dimension
        self.name_params_child = _as_py_name(
            self.file_name.replace("Properties", "")
        )

    def complete_params(self, params):
        try:
            constant_params = params["constant"]
        except AttributeError:
            constant_params = params._set_child("constant")

        _complete_params_dict(
            constant_params, self.name_params_child, self.default, self.doc
        )

    def make_tree(self, params):
        tree = FoamInputFile(
            info={
                "version": 2.0,
                "format": "ascii",
                "class": self.cls,
                "location": '"constant"',
                "object": self.file_name,
            },
            comments=self.comments,
        )

        if self.dimension is not None:
            tree.set_child("dimensions", DimensionSet(self.dimension))

        params_file = params.constant[self.name_params_child]

        default = deepcopy(self.default)
        _update_dict_with_params(default, params_file)

        tree.init_from_py_objects(
            default,
            dimensions=self.dimensions,
            default_dimension=self.default_dimension,
            comments=self.comments,
        )

        return tree



"""
Fluidsimfoam API reference

.. rubric:: Sub-packages

.. autosummary::
   :toctree:

   foam_input_files
   solvers
   output
   util
   resources

.. rubric:: Modules

.. autosummary::
   :toctree:

   info
   make
   init_fields
   operators
   log
   tasks
   testing
   params

"""
import importlib.metadata
import sys

from fluiddyn.io.redirect_stdout import stdout_redirected
from fluidsim_core.paths import find_path_result_dir
from fluidsimfoam.params import Parameters

__version__ = importlib.metadata.version(__package__ or __name__)
__all__ = ["load", "load_simul", "load_params"]


def load_simul(path_dir=".", hide_stdout=False):
    """Loads a simulation

    Parameters
    ----------
    path_dir: str or path-like
        Path to a directory containing a simulation. If not provided the
        current directory is used.

    """

    from fluidsimfoam.solvers import get_solver_short_name, import_cls_simul

    path_dir = find_path_result_dir(path_dir)

    short_name = get_solver_short_name(path_dir)
    Simul = import_cls_simul(short_name)

    path_params = path_dir / "params_simul.xml"
    if not path_params.exists():
        path_params = path_dir / ".data_fluidsim/params_simul.xml"

    params = Parameters(path_file=path_params)

    # Modify parameters prior to loading
    params.NEW_DIR_RESULTS = False
    params.output.HAS_TO_SAVE = False
    params.path_run = path_dir

    with stdout_redirected(hide_stdout):
        sim = Simul(params)

    return sim


def load_params(path_dir="."):
    """Load a :class:`fluidsim_core.params.Parameters` instance from ``path_dir``

    Parameters
    ----------
    path_dir : str or path-like
        Path to a simulation directory.

    Returns
    -------
    params: :class:`fluidsim_core.params.Parameters`

    """
    path_dir = find_path_result_dir(path_dir)
    return Parameters(path_file=path_dir / "params_simul.xml")


load = load_simul


if any("pytest" in part for part in sys.argv):
    import pytest

    pytest.register_assert_rewrite("fluidsimfoam.testing")
